import json
import logging
from typing import Any, Dict

from uaclient import exceptions, http, secret_manager, system, util
from uaclient.clouds import PublicCloudAutoAttachInstance

IMDS_IPV4_ADDRESS = "169.254.169.254"
IMDS_IPV6_ADDRESS = "[fd00:ec2::254]"

IMDS_IP_ADDRESS = (IMDS_IPV4_ADDRESS, IMDS_IPV6_ADDRESS)
IMDS_V2_TOKEN_URL = "http://{}/latest/api/token"
IMDS_URL = "http://{}/latest/dynamic/instance-identity/pkcs7"
_IMDS_IID_URL = "http://{}/latest/dynamic/instance-identity/document"

SYS_HYPERVISOR_PRODUCT_UUID = "/sys/hypervisor/uuid"
DMI_PRODUCT_SERIAL = "/sys/class/dmi/id/product_serial"
DMI_PRODUCT_UUID = "/sys/class/dmi/id/product_uuid"

AWS_TOKEN_TTL_SECONDS = "21600"
AWS_TOKEN_PUT_HEADER = "X-aws-ec2-metadata-token"
AWS_TOKEN_REQ_HEADER = AWS_TOKEN_PUT_HEADER + "-ttl-seconds"

LOG = logging.getLogger(util.replace_top_level_logger_name(__name__))


class AWSAutoAttachInstance(PublicCloudAutoAttachInstance):
    _api_token = None
    _ip_address = None

    def _get_imds_url_response(self, url: str, headers):
        response = http.readurl(url, headers=headers, timeout=1)
        if response.code == 200:
            return response.body
        else:
            raise exceptions.CloudMetadataError(
                code=response.code, body=response.body
            )

    # mypy does not handle @property around inner decorators
    # https://github.com/python/mypy/issues/1362
    @property  # type: ignore
    @util.retry(exceptions.CloudMetadataError, retry_sleeps=[0.5, 1, 1])
    def identity_doc(self) -> Dict[str, Any]:
        headers = self._request_imds_v2_token_headers()
        url = IMDS_URL.format(self._ip_address)
        imds_url_response = self._get_imds_url_response(url, headers=headers)
        secret_manager.secrets.add_secret(imds_url_response)
        return {"pkcs7": imds_url_response}

    @util.retry(exceptions.CloudMetadataError, retry_sleeps=[0.5, 1, 1])
    def _get_ii_doc(self) -> Dict:
        """
        Get the instance identity doc associated with the current instance.

        See
        https://docs.aws.amazon.com/AWSEC2/latest/UserGuide/retrieve-iid.html
        for more context.

        @return: Dict containing the instance identity document.
        """
        headers = self._request_imds_v2_token_headers()
        url = _IMDS_IID_URL.format(self._ip_address)
        try:
            ii_doc = json.loads(
                self._get_imds_url_response(url, headers=headers)
            )
        except json.JSONDecodeError as e:
            LOG.debug("Error decoding instance identity document: %s", e)
            return {}
        return ii_doc

    @property  # type: ignore
    def is_likely_pro(self) -> bool:
        """
        Determines if the instance is likely Ubuntu Pro.

        Criteria: if any billing-product or marketplace-product-code is
        present, then is likely a Pro instance.

        @return: Boolean indicating if the instance is likely pro or not.
        """
        ii_doc = self._get_ii_doc()
        billing_products = ii_doc.get("billingProducts", None)
        marketplace_product_codes = ii_doc.get("marketplaceProductCodes", None)
        return bool(billing_products) or bool(marketplace_product_codes)

    def _request_imds_v2_token_headers(self):
        for address in IMDS_IP_ADDRESS:
            try:
                headers = self._get_imds_v2_token_headers(ip_address=address)
            except Exception as e:
                LOG.warning(
                    "Could not reach AWS IMDS at http://%s: %s\n",
                    address,
                    getattr(e, "reason", ""),
                )
            else:
                self._ip_address = address
                break
        if self._ip_address is None:
            raise exceptions.AWSNoValidIMDS(
                addresses=", ".join(IMDS_IP_ADDRESS)
            )
        return headers

    @util.retry(exceptions.CloudMetadataError, retry_sleeps=[1, 2, 5])
    def _get_imds_v2_token_headers(self, ip_address):
        if self._api_token == "IMDSv1":
            return None
        elif self._api_token:
            return {AWS_TOKEN_PUT_HEADER: self._api_token}

        response = http.readurl(
            IMDS_V2_TOKEN_URL.format(ip_address),
            method="PUT",
            headers={AWS_TOKEN_REQ_HEADER: AWS_TOKEN_TTL_SECONDS},
            timeout=1,
        )
        if response.code == 200:
            self._api_token = response.body
            secret_manager.secrets.add_secret(self._api_token)
            return {AWS_TOKEN_PUT_HEADER: self._api_token}
        if response.code == 404:
            self._api_token = "IMDSv1"
            return None

        raise exceptions.CloudMetadataError(
            code=response.code, body=response.body
        )

    @property
    def cloud_type(self) -> str:
        return "aws"

    @property
    def is_viable(self) -> bool:
        """This machine is a viable AWSInstance"""
        try:
            hypervisor_uuid = system.load_file(SYS_HYPERVISOR_PRODUCT_UUID)
            if "ec2" == hypervisor_uuid[0:3]:
                return True
        except FileNotFoundError:
            # SYS_HYPERVISOR_PRODUCT_UUID isn't present on all EC2 instance
            # types, fall through
            pass
        # Both DMI product_uuid and product_serial start with 'ec2'
        dmi_uuid = system.load_file(DMI_PRODUCT_UUID).lower()
        dmi_serial = system.load_file(DMI_PRODUCT_SERIAL).lower()
        if "ec2" == dmi_uuid[0:3] == dmi_serial[0:3]:
            return True
        return False

    def should_poll_for_pro_license(self) -> bool:
        """Unsupported"""
        return False

    def is_pro_license_present(self, *, wait_for_change: bool) -> bool:
        raise exceptions.InPlaceUpgradeNotSupportedError()



import json

from behave import then, when

from features.steps.shell import when_i_run_command
from features.util import SUT


@when("I do a preflight check for `{contract_token}` {user_spec}")
def when_i_preflight(context, contract_token, user_spec, verify_return=True):
    token = getattr(context.pro_config, contract_token, "invalid_token")
    command = "pro status --simulate-with-token {}".format(token)
    if user_spec == "with the all flag":
        command += " --all"
    if "formatted as" in user_spec:
        output_format = user_spec.split()[2]
        command += " --format {}".format(output_format)
    when_i_run_command(
        context=context,
        command=command,
        user_spec="as non-root",
        verify_return=verify_return,
    )


@when(
    "I verify that a preflight check for `{contract_token}` {user_spec} exits {exit_codes}"  # noqa
)
def when_i_attempt_preflight(context, contract_token, user_spec, exit_codes):
    when_i_preflight(context, contract_token, user_spec, verify_return=False)

    expected_codes = exit_codes.split(",")
    assert str(context.process.returncode) in expected_codes


def get_enabled_services(context, machine_name=SUT):
    when_i_run_command(
        context,
        "pro api u.pro.status.enabled_services.v1",
        "as non-root",
        machine_name=machine_name,
    )

    data = json.loads(context.process.stdout.strip())

    enabled_services = []
    warning_services = []
    for enabled_service in data["data"]["attributes"]["enabled_services"]:
        if enabled_service["variant_enabled"]:
            enabled_services.append(enabled_service["variant_name"])
        else:
            enabled_services.append(enabled_service["name"])

    for warning in data["warnings"]:
        warning_services.append(warning["meta"]["service"])

    return enabled_services, warning_services


@then("I verify that `{service}` is disabled")
def i_verify_that_service_is_disabled(context, service):
    enabled_services, _ = get_enabled_services(context)

    if service in enabled_services:
        raise AssertionError(
            "Expected {} to not be enabled\nEnabled services: {}".format(
                service, ", ".join(enabled_services)
            )
        )


@then("I verify that `{service}` is enabled")
def i_verify_that_service_is_enabled(context, service):
    enabled_services, _ = get_enabled_services(context)

    if service not in enabled_services:
        raise AssertionError(
            "Expected {} to be enabled\nEnabled services: {}".format(
                service, ", ".join(enabled_services)
            )
        )


@then("I verify that `{service}` status is warning")
def i_verify_that_service_status_is_warning(context, service):
    enabled_services, warning_services = get_enabled_services(context)

    if service not in enabled_services:
        msg = (
            "Expected {} status to be warning, but the service is disabled\n"
            "Enabled services: {}"
        )
        raise AssertionError(msg.format(service, ", ".join(enabled_services)))

    if service not in warning_services:
        msg = "Expected {} status to be warning, but the status is enabled"
        raise AssertionError(msg.format(service))


@then("I verify that `{service}` status is `{status}`")
def i_verify_service_status(context, service, status):
    if status == "enabled":
        i_verify_that_service_is_enabled(context, service)
    elif status == "disabled":
        i_verify_that_service_is_disabled(context, service)
    elif status == "warning":
        i_verify_that_service_status_is_warning(context, service)
    else:
        raise AssertionError(
            "Service status {} is not supported".format(status)
        )



import socket
import urllib
from urllib.parse import urlparse

import mock
import pytest

from uaclient import exceptions, http, messages


class TestIsServiceUrl:
    @pytest.mark.parametrize(
        "url,is_valid",
        (
            ("http://asdf", True),
            ("http://asdf/", True),
            ("asdf", False),
            ("http://host:port", False),
            ("http://asdf:1234", True),
        ),
    )
    def test_is_valid_url(self, url, is_valid):
        ret = http.is_service_url(url)
        assert is_valid is ret


class TestValidateProxy:
    @pytest.mark.parametrize(
        "proxy", ["invalidurl", "htp://wrongscheme", "http//missingcolon"]
    )
    @mock.patch("urllib.request.OpenerDirector.open")
    def test_fails_on_invalid_url(self, m_open, proxy):
        """
        Check that invalid urls are rejected with the correct message
        and that we don't even attempt to use them
        """
        with pytest.raises(exceptions.UbuntuProError) as e:
            http.validate_proxy("http", proxy, "http://example.com")

        assert (
            e.value.msg
            == messages.E_NOT_SETTING_PROXY_INVALID_URL.format(proxy=proxy).msg
        )

    @pytest.mark.parametrize(
        "protocol, proxy, test_url",
        [
            ("http", "http://localhost:1234", "http://example.com"),
            ("https", "http://localhost:1234", "https://example.com"),
            ("https", "https://localhost:1234", "https://example.com"),
        ],
    )
    @mock.patch("urllib.request.Request")
    @mock.patch("urllib.request.ProxyHandler")
    @mock.patch("urllib.request.build_opener")
    @mock.patch("urllib.request.OpenerDirector.open")
    @mock.patch("uaclient.http._readurl_pycurl_https_in_https")
    def test_calls_open_on_valid_url(
        self,
        m_readurl_pycurl,
        m_open,
        m_build_opener,
        m_proxy_handler,
        m_request,
        protocol,
        proxy,
        test_url,
    ):
        """
        Check that we attempt to use a valid url as a proxy
        Also check that we return the proxy value when the open call succeeds
        """
        m_request_resp = mock.MagicMock()
        m_build_opener.return_value = urllib.request.OpenerDirector()
        m_request.return_value = m_request_resp
        m_readurl_pycurl.return_value = mock.MagicMock(code=200)

        ret = http.validate_proxy(protocol, proxy, test_url)

        assert [mock.call(test_url, method="HEAD")] == m_request.call_args_list

        if protocol == "https" and urlparse(proxy).scheme == "https":
            assert [
                mock.call(m_request_resp, https_proxy=proxy)
            ] == m_readurl_pycurl.call_args_list
        else:
            assert [
                mock.call({protocol: proxy})
            ] == m_proxy_handler.call_args_list
            assert 1 == m_build_opener.call_count
            assert 1 == m_open.call_count

        assert proxy == ret

    @pytest.mark.parametrize(
        "open_side_effect, expected_message",
        [
            (socket.timeout(0, "timeout"), "[Errno 0] timeout"),
            (urllib.error.URLError("reason"), "reason"),
        ],
    )
    @mock.patch("urllib.request.OpenerDirector.open")
    def test_fails_when_open_fails(
        self, m_open, open_side_effect, expected_message, caplog_text
    ):
        """
        Check that we return the appropriate error when the proxy doesn't work
        """
        m_open.side_effect = open_side_effect
        with pytest.raises(exceptions.UbuntuProError) as e:
            http.validate_proxy(
                "http", "http://localhost:1234", "http://example.com"
            )

        assert (
            e.value.msg
            == messages.E_NOT_SETTING_PROXY_NOT_WORKING.format(
                proxy="http://localhost:1234"
            ).msg
        )

        assert (
            'Error trying to use "http://localhost:1234" as urllib proxy '
            'to reach "http://example.com": {}'
        ).format(expected_message) in caplog_text()


class TestConfigureWebProxy:
    @pytest.mark.parametrize(
        "http_proxy,https_proxy,m_environ,expected_environ",
        (
            (
                None,
                None,
                {},
                {
                    "NO_PROXY": "169.254.169.254,[fd00:ec2::254],metadata",
                    "no_proxy": "169.254.169.254,[fd00:ec2::254],metadata",
                },
            ),
            (
                "http://proxy",
                "https://proxy",
                {"no_proxy": "a,10.0.0.1"},
                {
                    "NO_PROXY": "10.0.0.1,169.254.169.254,[fd00:ec2::254],a,metadata",  # noqa
                    "no_proxy": "10.0.0.1,169.254.169.254,[fd00:ec2::254],a,metadata",  # noqa
                },
            ),
            (
                "http://proxy",
                "https://proxy",
                {"NO_PROXY": "a,169.254.169.254"},
                {
                    "NO_PROXY": "169.254.169.254,[fd00:ec2::254],a,metadata",
                    "no_proxy": "169.254.169.254,[fd00:ec2::254],a,metadata",
                },
            ),
        ),
    )
    @mock.patch("urllib.request.OpenerDirector.open")
    def test_no_proxy_set_in_environ(
        self, m_open, http_proxy, https_proxy, m_environ, expected_environ
    ):
        with mock.patch.dict(http.os.environ, m_environ, clear=True):
            http.configure_web_proxy(
                http_proxy=http_proxy, https_proxy=https_proxy
            )
            assert expected_environ == http.os.environ


def dict_eq(self, other):
    return self.__dict__ == other.__dict__


def dict_repr(self):
    return (
        "{\n  "
        + "\n  ".join(
            [
                "{}: {}".format(repr(k), repr(v))
                for k, v in self.__dict__.items()
            ]
        )
        + "\n}"
    )


@mock.patch.object(urllib.request.Request, "__eq__", dict_eq)
@mock.patch.object(urllib.request.Request, "__repr__", dict_repr)
class TestReadurl:
    @pytest.mark.parametrize(
        [
            "url",
            "data",
            "headers",
            "method",
            "timeout",
            "expected_urllib_calls",
            "urllib_return_value",
            "expected_response",
        ],
        [
            # simplest GET
            (
                "http://example.com",
                None,
                {},
                None,
                None,
                [
                    mock.call(
                        urllib.request.Request(
                            "http://example.com",
                            data=None,
                            headers={},
                            method=None,
                        ),
                        timeout=None,
                    )
                ],
                http.UnparsedHTTPResponse(
                    code=200,
                    headers={"something": "here"},
                    body=b"body",
                ),
                http.HTTPResponse(
                    code=200,
                    headers={"something": "here"},
                    body="body",
                    json_dict={},
                    json_list=[],
                ),
            ),
            # passes through timeout, headers, method
            (
                "http://example.com",
                None,
                {"req": "header"},
                "PUT",
                1,
                [
                    mock.call(
                        urllib.request.Request(
                            "http://example.com",
                            data=None,
                            headers={"req": "header"},
                            method="PUT",
                        ),
                        timeout=1,
                    )
                ],
                http.UnparsedHTTPResponse(
                    code=200,
                    headers={"something": "here"},
                    body=b"body",
                ),
                http.HTTPResponse(
                    code=200,
                    headers={"something": "here"},
                    body="body",
                    json_dict={},
                    json_list=[],
                ),
            ),
            # implicit POST with data
            (
                "http://example.com",
                b"data",
                {},
                None,
                None,
                [
                    mock.call(
                        urllib.request.Request(
                            "http://example.com",
                            data=b"data",
                            headers={},
                            method="POST",
                        ),
                        timeout=None,
                    )
                ],
                http.UnparsedHTTPResponse(
                    code=200,
                    headers={"something": "here"},
                    body=b"body",
                ),
                http.HTTPResponse(
                    code=200,
                    headers={"something": "here"},
                    body="body",
                    json_dict={},
                    json_list=[],
                ),
            ),
            # override method with data
            (
                "http://example.com",
                b"data",
                {},
                "PATCH",
                None,
                [
                    mock.call(
                        urllib.request.Request(
                            "http://example.com",
                            data=b"data",
                            headers={},
                            method="PATCH",
                        ),
                        timeout=None,
                    )
                ],
                http.UnparsedHTTPResponse(
                    code=200,
                    headers={"something": "here"},
                    body=b"body",
                ),
                http.HTTPResponse(
                    code=200,
                    headers={"something": "here"},
                    body="body",
                    json_dict={},
                    json_list=[],
                ),
            ),
            # response json dict parsed
            (
                "http://example.com",
                None,
                {},
                None,
                None,
                [
                    mock.call(
                        urllib.request.Request(
                            "http://example.com",
                            data=None,
                            headers={},
                            method=None,
                        ),
                        timeout=None,
                    )
                ],
                http.UnparsedHTTPResponse(
                    code=200,
                    headers={"content-type": "application/json"},
                    body=b'{"hello": "hello"}',
                ),
                http.HTTPResponse(
                    code=200,
                    headers={"content-type": "application/json"},
                    body='{"hello": "hello"}',
                    json_dict={"hello": "hello"},
                    json_list=[],
                ),
            ),
            # response json list parsed
            (
                "http://example.com",
                None,
                {},
                None,
                None,
                [
                    mock.call(
                        urllib.request.Request(
                            "http://example.com",
                            data=None,
                            headers={},
                            method=None,
                        ),
                        timeout=None,
                    )
                ],
                http.UnparsedHTTPResponse(
                    code=200,
                    headers={"content-type": "application/json"},
                    body=b'["hello"]',
                ),
                http.HTTPResponse(
                    code=200,
                    headers={"content-type": "application/json"},
                    body='["hello"]',
                    json_dict={},
                    json_list=["hello"],
                ),
            ),
        ],
    )
    @mock.patch("uaclient.http._readurl_urllib")
    def test_readurl(
        self,
        m_readurl_urllib,
        url,
        data,
        headers,
        method,
        timeout,
        expected_urllib_calls,
        urllib_return_value,
        expected_response,
    ):
        # urllib.request.Request.__eq__ == _RequestPatch.__eq__
        m_readurl_urllib.return_value = urllib_return_value
        assert expected_response == http.readurl(
            url, data, headers, method, timeout
        )
        assert expected_urllib_calls == m_readurl_urllib.call_args_list


class TestShouldUsePycurl:
    @pytest.mark.parametrize("proxy_bypass", ((True), (False)))
    @pytest.mark.parametrize(
        "https_proxy,target_url,expected_return",
        (
            ("https://proxy:443", "https://www.test.com", True),
            ("http://proxy:443", "https://www.test.com", False),
            ("https://proxy:443", "http://www.test.com", False),
        ),
    )
    @mock.patch("urllib.request.proxy_bypass")
    def test_should_use_pycurl(
        self,
        m_proxy_bypass,
        https_proxy,
        target_url,
        expected_return,
        proxy_bypass,
    ):
        m_proxy_bypass.return_value = proxy_bypass

        if proxy_bypass:
            assert not http.should_use_pycurl(https_proxy, target_url)
        else:
            assert expected_return == http.should_use_pycurl(
                https_proxy, target_url
            )


class TestHandlePycurlError:
    @pytest.mark.parametrize(
        "error_args,expected_exception",
        (
            (("PYCURL_ERROR", "test"), exceptions.PycurlError),
            (("NON_PYCURL_ERROR", "test"), exceptions.PycurlError),
            (
                ("PYCURL_ERROR", "HTTP code 407 from proxy: proxy"),
                exceptions.ProxyAuthenticationFailed,
            ),
            (("PYCURL_ERROR", "test"), exceptions.PycurlError),
            (
                ("CA_CERTIFICATES_ERROR", "test"),
                exceptions.PycurlCACertificatesError,
            ),
        ),
    )
    def test_handle_pycurl_error(self, error_args, expected_exception):
        with pytest.raises(expected_exception):
            m_error = mock.MagicMock(args=error_args)
            http._handle_pycurl_error(
                m_error, "url", "PYCURL_ERROR", "CA_CERTIFICATES_ERROR"
            )


class TestUnixSocketRequest:
    @pytest.mark.parametrize(
        [
            "params",
            "response_body",
            "response_code",
            "response_headers",
            "expected",
            "sock_connect_call_args",
            "http_connection_call_args",
            "request_call_args",
        ],
        [
            (
                {
                    "socket_path": "/fake/socket",
                    "http_method": "GET",
                    "http_path": "/fake/path",
                    "http_hostname": "x",
                },
                b"body",
                200,
                {"header": "fake"},
                http.HTTPResponse(
                    code=200,
                    headers={"header": "fake"},
                    body="body",
                    json_dict={},
                    json_list=[],
                ),
                [mock.call.connect("/fake/socket")],
                [mock.call("x")],
                [mock.call("GET", "/fake/path")],
            ),
            (
                {
                    "socket_path": "/fake/socket2",
                    "http_method": "POST",
                    "http_path": "/fake/path2",
                    "http_hostname": "x2",
                },
                b"body",
                201,
                {"header2": "fake2"},
                http.HTTPResponse(
                    code=201,
                    headers={"header2": "fake2"},
                    body="body",
                    json_dict={},
                    json_list=[],
                ),
                [mock.call.connect("/fake/socket2")],
                [mock.call("x2")],
                [mock.call("POST", "/fake/path2")],
            ),
            (
                {
                    "socket_path": "/fake/socket",
                    "http_method": "GET",
                    "http_path": "/fake/path",
                    "http_hostname": "x",
                },
                b'["body"]',
                200,
                {"header": "fake"},
                http.HTTPResponse(
                    code=200,
                    headers={"header": "fake"},
                    body='["body"]',
                    json_dict={},
                    json_list=[],
                ),
                [mock.call.connect("/fake/socket")],
                [mock.call("x")],
                [mock.call("GET", "/fake/path")],
            ),
            (
                {
                    "socket_path": "/fake/socket",
                    "http_method": "GET",
                    "http_path": "/fake/path",
                    "http_hostname": "x",
                },
                b'["body"]',
                200,
                {"content-type": "application/json"},
                http.HTTPResponse(
                    code=200,
                    headers={"content-type": "application/json"},
                    body='["body"]',
                    json_dict={},
                    json_list=["body"],
                ),
                [mock.call.connect("/fake/socket")],
                [mock.call("x")],
                [mock.call("GET", "/fake/path")],
            ),
            (
                {
                    "socket_path": "/fake/socket",
                    "http_method": "GET",
                    "http_path": "/fake/path",
                    "http_hostname": "x",
                },
                b'{"body": "value"}',
                200,
                {"content-type": "application/json"},
                http.HTTPResponse(
                    code=200,
                    headers={"content-type": "application/json"},
                    body='{"body": "value"}',
                    json_dict={"body": "value"},
                    json_list=[],
                ),
                [mock.call.connect("/fake/socket")],
                [mock.call("x")],
                [mock.call("GET", "/fake/path")],
            ),
        ],
    )
    @mock.patch("uaclient.http.http.client.HTTPConnection")
    @mock.patch("uaclient.http.socket.socket")
    def test_unix_socket_request(
        self,
        m_socket,
        m_http_connection,
        params,
        response_body,
        response_code,
        response_headers,
        expected,
        sock_connect_call_args,
        http_connection_call_args,
        request_call_args,
    ):
        conn = m_http_connection.return_value
        conn.getresponse.return_value.read.return_value = response_body
        conn.getresponse.return_value.status = response_code
        conn.getresponse.return_value.headers = response_headers

        assert expected == http.unix_socket_request(**params)

        assert m_socket.called
        assert (
            sock_connect_call_args
            == m_socket.return_value.connect.call_args_list
        )
        assert http_connection_call_args == m_http_connection.call_args_list
        assert conn.sock == m_socket.return_value
        assert request_call_args == conn.request.call_args_list
        assert m_socket.return_value.close.called
        assert conn.close.called



"""
This module contains URLs that are to be displayed in messages only.
URLs that are actually contacted by the pro-client should be defined elsewhere.
"""

PRO_HOME_PAGE = "https://ubuntu.com/pro"
PRO_DASHBOARD = "https://ubuntu.com/pro/dashboard"
PRO_ATTACH = "https://ubuntu.com/pro/attach"
PRO_SUBSCRIBE = "https://ubuntu.com/pro/subscribe"

PRO_CLIENT_DOCS_RELATED_USNS = "https://canonical-ubuntu-pro-client.readthedocs-hosted.com/en/latest/explanations/cves_and_usns_explained.html#what-are-related-usns"  # noqa: E501
PRO_CLIENT_DOCS_CLOUD_PRO_IMAGES = "https://canonical-ubuntu-pro-client.readthedocs-hosted.com/en/latest/explanations/what_are_ubuntu_pro_cloud_instances.html"  # noqa: E501
PRO_CLIENT_DOCS_PROXY_CONFIG = "https://canonical-ubuntu-pro-client.readthedocs-hosted.com/en/latest/howtoguides/configure_proxies.html"  # noqa: E501
PRO_CLIENT_DOCS_API_REFERENCE = "https://canonical-ubuntu-pro-client.readthedocs-hosted.com/en/latest/references/api/"  # noqa: E501

# TODO: If/when `pro disable fips --purge` exists, reference that where
# this URL is used
# If that doesn't happen, write a new how-to-guide in our proper docs and
# link to that here
PRO_CLIENT_DOCS_REMOVE_FIPS = "https://discourse.ubuntu.com/t/20738"

PRO_ON_AWS_HOME_PAGE = "https://ubuntu.com/aws/pro"
PRO_ON_AZURE_HOME_PAGE = "https://ubuntu.com/azure/pro"
PRO_ON_GCP_HOME_PAGE = "https://ubuntu.com/gcp/pro"

ANBOX_HOME_PAGE = "https://anbox-cloud.io"
ANBOX_DOCS_APPLIANCE_INITIALIZE = "https://anbox-cloud.io/docs/tutorial/installing-appliance/#initialize-the-appliance"  # noqa: E501
CIS_HOME_PAGE = "https://ubuntu.com/security/cis"
COMMON_CRITERIA_HOME_PAGE = "https://ubuntu.com/security/cc"
ESM_HOME_PAGE = "https://ubuntu.com/security/esm"
FIPS_HOME_PAGE = "https://ubuntu.com/security/fips"
LANDSCAPE_HOME_PAGE = "https://ubuntu.com/landscape"
LANDSCAPE_SAAS = "https://landscape.canonical.com"
LANDSCAPE_DOCS_INSTALL = "https://ubuntu.com/landscape/install"
LANDSCAPE_CLIENT_INSTALL = (
    "https://ubuntu.com/landscape/docs/install-landscape-client"
)
LIVEPATCH_HOME_PAGE = "https://ubuntu.com/security/livepatch"
LIVEPATCH_SUPPORTED_KERNELS = (
    "https://ubuntu.com/security/livepatch/docs/kernels"
)
REALTIME_HOME_PAGE = "https://ubuntu.com/realtime-kernel"
ROS_HOME_PAGE = "https://ubuntu.com/robotics/ros-esm"
USG_DOCS = "https://ubuntu.com/security/certifications/docs/usg"

SECURITY_CVE_PAGE = "https://ubuntu.com/security/{cve}"

GCP_SERVICE_ACCOUNT_DOCS = (
    "https://cloud.google.com/iam/docs/service-account-overview"
)



import os
from typing import Tuple, Type, Union

from uaclient import api, messages, system
from uaclient.apt import APT_KEYS_DIR, DEB822_REPO_FILE_CONTENT, KEYRINGS_DIR
from uaclient.defaults import ESM_APT_ROOTDIR
from uaclient.entitlements import repo
from uaclient.entitlements.base import UAEntitlement
from uaclient.entitlements.entitlement_status import CanDisableFailure
from uaclient.util import set_filename_extension


class ESMBaseEntitlement(repo.RepoEntitlement):
    help_doc_url = messages.urls.ESM_HOME_PAGE
    check_updates_pocket = True

    @property
    def dependent_services(self) -> Tuple[Type[UAEntitlement], ...]:
        from uaclient.entitlements.ros import (
            ROSEntitlement,
            ROSUpdatesEntitlement,
        )

        return (ROSUpdatesEntitlement, ROSEntitlement)

    def _perform_enable(self, progress: api.ProgressWrapper) -> bool:
        from uaclient.timer.update_messaging import update_motd_messages

        enable_performed = super()._perform_enable(progress)
        if enable_performed:
            update_motd_messages(self.cfg)
            self.disable_local_esm_repo()
        return enable_performed

    def setup_local_esm_repo(self) -> None:
        series = system.get_release_info().series
        # Ugly? Yes, but so is python < 3.8 without removeprefix
        assert self.name.startswith("esm-")
        esm_name = self.name[len("esm-") :]
        sources_repo_filename = set_filename_extension(
            os.path.normpath(
                ESM_APT_ROOTDIR + self.repo_file,
            ),
            "sources",
        )
        list_repo_filename = set_filename_extension(
            os.path.normpath(
                ESM_APT_ROOTDIR + self.repo_file,
            ),
            "list",
        )

        # No need to create if any format already present
        if os.path.exists(sources_repo_filename) or os.path.exists(
            list_repo_filename
        ):
            return

        esm_url = "https://esm.ubuntu.com/{name}/ubuntu".format(name=esm_name)
        suites = "{series}-{name}-security {series}-{name}-updates".format(
            series=series, name=esm_name
        )

        # When writing, use the sources format by default
        system.write_file(
            sources_repo_filename,
            DEB822_REPO_FILE_CONTENT.format(
                url=esm_url,
                suites=suites,
                keyrings_dir=KEYRINGS_DIR,
                keyring_file=self.repo_key_file,
                deb_src="",
            ),
        )

    def disable_local_esm_repo(self) -> None:
        keyring_file = os.path.normpath(
            ESM_APT_ROOTDIR + APT_KEYS_DIR + self.repo_key_file
        )
        system.ensure_file_absent(keyring_file)

        repo_filename = os.path.normpath(
            ESM_APT_ROOTDIR + self.repo_file,
        )
        # Remove any instance of the file present in the folder
        system.ensure_file_absent(
            set_filename_extension(repo_filename, "sources")
        )
        system.ensure_file_absent(
            set_filename_extension(repo_filename, "list")
        )


class ESMAppsEntitlement(ESMBaseEntitlement):
    origin = "UbuntuESMApps"
    name = "esm-apps"
    title = messages.ESM_APPS_TITLE
    description = messages.ESM_APPS_DESCRIPTION
    help_text = messages.ESM_APPS_HELP_TEXT
    repo_key_file = "ubuntu-pro-esm-apps.gpg"

    def disable(
        self, progress: api.ProgressWrapper
    ) -> Tuple[bool, Union[None, CanDisableFailure]]:
        from uaclient.timer.update_messaging import update_motd_messages

        disable_performed, fail = super().disable(progress)
        if disable_performed:
            update_motd_messages(self.cfg)
            if system.is_current_series_lts():
                self.setup_local_esm_repo()
        return disable_performed, fail


class ESMInfraEntitlement(ESMBaseEntitlement):
    name = "esm-infra"
    origin = "UbuntuESM"
    title = messages.ESM_INFRA_TITLE
    description = messages.ESM_INFRA_DESCRIPTION
    help_text = messages.ESM_INFRA_HELP_TEXT
    repo_key_file = "ubuntu-pro-esm-infra.gpg"

    def disable(
        self, progress: api.ProgressWrapper
    ) -> Tuple[bool, Union[None, CanDisableFailure]]:
        from uaclient.timer.update_messaging import update_motd_messages

        disable_performed, fail = super().disable(progress)
        if disable_performed:
            update_motd_messages(self.cfg)
            if system.is_current_series_active_esm():
                self.setup_local_esm_repo()
        return disable_performed, fail



#!/usr/bin/python3
"""
Generate documentation for an ubuntu-pro-client cli commands.
"""

import sys

sys.path.insert(0, ".")

from uaclient.cli import COMMANDS, get_parser
from uaclient.cli.commands import ProCommand

VALID_TARGETS = ["manpage", "rst"]

GENERATED_DOC_HEADER = """\
..
   THIS DOCUMENTATION WAS AUTOMATICALLY GENERATED
   Do not edit this document directly. Instead, edit the commands in the cli
   source code, which can be found on the main branch of the git repo at
   https://github.com/canonical/ubuntu-pro-client. The cli source code is
   nested in the uaclient/cli folder. Description for the commands is found
   in the uaclient/messages folder.

Available commands
==================

The currently available commands are:

{commands_list}

"""
COMMAND_LINK_TEMPLATE = "- `{command_entry}`_\n"

RST_TEMPLATE = """\
{command}
{section_mark}

**Usage:**

``{usage_line}``

**Description:**

{description}

"""

MANPAGE_TEMPLATE = """\
.TP
.BR "{indent}{name}" " {usage_options}"
{description}

"""


def _build_rst_entry(command: ProCommand, section_mark: str = "="):
    return RST_TEMPLATE.format(
        command=command.parser.prog,
        section_mark=section_mark * len(command.parser.prog),
        usage_line=" ".join(
            command.parser.format_usage()[len("usage: ") :].split()
        ),
        description=command.description,
    )


def _build_manpage_entry(command: ProCommand, indent: str = "") -> str:
    # the usage line looks like
    # usage: <prog> <options>
    # we want only the <options>
    options_line = command.parser.format_usage().split(command.parser.prog)[1]
    # Remove extra spaces and eventual newlines
    usage_options = " ".join(options_line.split())

    return MANPAGE_TEMPLATE.format(
        indent=indent,
        name=command.name,
        usage_options=usage_options,
        description=command.description,
    )


def _print_rst_section():
    command_list = ""
    for command in COMMANDS:
        command_list += COMMAND_LINK_TEMPLATE.format(
            command_entry=command.parser.prog
        )

    content = GENERATED_DOC_HEADER.format(commands_list=command_list)
    for command in COMMANDS:
        content += _build_rst_entry(command)
        for subcommand in command.subcommands:
            content += _build_rst_entry(subcommand, section_mark="-")

    print(content)


def _generate_manpage_section():
    result = ""
    for command in COMMANDS:
        result += _build_manpage_entry(command)
        for subcommand in command.subcommands:
            result += _build_manpage_entry(subcommand, indent=" " * 4)

    with open("./ubuntu-advantage.1.template", "r") as f:
        template = f.read()
    content = template.replace("<<commands_description>>", result)
    with open("./ubuntu-advantage.1", "w") as f:
        f.write(content)


if __name__ == "__main__":
    # get a parser so we register all the commands
    _parser = get_parser()
    if len(sys.argv) > 1:
        if sys.argv[1] == "manpage":
            _generate_manpage_section()
            sys.exit()
        if sys.argv[1] == "rst":
            _print_rst_section()
            sys.exit()

    raise ValueError("call the script with one of: " + " ".join(VALID_TARGETS))



import os
import stat

import mock

from uaclient import system
from uaclient.files import MachineTokenFile, UAFile


class TestUAFile:
    def test_read_write(self, tmpdir):
        file_name = "temp_file"
        file = UAFile(file_name, tmpdir.strpath, False)
        content = "dummy file words"
        file.write(content)
        path = os.path.join(tmpdir.strpath, file_name)
        res = system.load_file(path)
        assert res == file.read()
        assert res == content

    def test_non_private_file_world_readable(
        self,
        tmpdir,
    ):
        file = UAFile("test", tmpdir.strpath, False)
        file.write("dummy file words")

        assert 0o644 == stat.S_IMODE(
            os.lstat(os.path.join(tmpdir.strpath, "test")).st_mode
        )


class TestMachineTokenFile:
    def test_deleting(self, tmpdir):
        token_file = MachineTokenFile(
            directory=tmpdir.strpath,
        )
        token = {"machineTokenInfo": {"machineId": "random-id"}}
        token_file.write(token)
        assert token_file.machine_token == token
        token_file.delete()
        assert token_file.machine_token is None

    @mock.patch("uaclient.util.we_are_currently_root")
    def test_public_file_filtering(self, m_we_are_currently_root, tmpdir):
        # root access of machine token file
        m_we_are_currently_root.return_value = True
        token_file = MachineTokenFile(
            directory=tmpdir.strpath,
        )
        token = {
            "machineTokenInfo": {"machineId": "random-id"},
            "machineToken": "token",
        }
        token_file.write(token)
        root_token = token_file.machine_token
        assert token == root_token
        # non root access of machine token file
        m_we_are_currently_root.return_value = False
        token_file = MachineTokenFile(directory=tmpdir.strpath)
        nonroot_token = token_file.machine_token
        assert root_token != nonroot_token
        machine_token = nonroot_token.get("machineToken", None)
        assert machine_token is None



import json
import os
import tempfile

import yaml
from behave import then, when
from hamcrest import assert_that, equal_to, matches_regexp

from features.steps.shell import when_i_run_command
from features.util import SUT, process_template_vars
from uaclient.defaults import DEFAULT_CONFIG_FILE


def _get_file_contents(
    context, file_path: str, machine_name: str = SUT
) -> str:
    # pycloudlib pull_file doesn't work on root readable files
    when_i_run_command(
        context,
        "cat {}".format(file_path),
        "with sudo",
        machine_name=machine_name,
    )
    return context.process.stdout


@when("I push static file `{file_name}` to machine")
def when_i_push_file(context, file_name, machine_name=SUT):
    context.machines[machine_name].instance.push_file(
        os.path.join("features/files/", file_name), "/tmp/{}".format(file_name)
    )


@when("I add this text on `{file_name}` on `{machine_name}` above `{line}`:")
def when_i_add_this_text_on_file_above_line(
    context, file_name, machine_name, line
):
    command = 'sed -i "s/{}/{}\\n{}/" {}'.format(
        line, context.text, line, file_name
    )
    when_i_run_command(
        context, command, "with sudo", machine_name=machine_name
    )


@when("I verify `{file_name}` is empty")
@when("I verify `{file_name}` is empty on `{machine_name}` machine")
def when_i_verify_file_is_empty_on_machine(
    context, file_name, machine_name=SUT
):
    command = 'sh -c "cat {} | wc -l"'.format(file_name)
    when_i_run_command(
        context, command, user_spec="with sudo", machine_name=machine_name
    )

    assert_that(context.process.stdout.strip(), matches_regexp("0"))


@when(
    "I create the file `{file_path}` on the `{machine_name}` machine with the following:"  # noqa: E501
)
@when("I create the file `{file_path}` with the following:")
def when_i_create_file_with_content(
    context, file_path, machine_name=SUT, text=None
):
    if text is None:
        text = context.text
    text = process_template_vars(context, text)

    with tempfile.TemporaryDirectory() as tmpd:
        tmpf_path = os.path.join(tmpd, "tmpfile")
        with open(tmpf_path, mode="w") as tmpf:
            tmpf.write(text)
        context.machines[machine_name].instance.push_file(
            tmpf_path, "/tmp/behave_tmpfile"
        )
    dirname = os.path.dirname(file_path)
    if dirname:
        when_i_run_command(
            context,
            "mkdir -p {}".format(dirname),
            "with sudo",
            machine_name=machine_name,
        )
    when_i_run_command(
        context,
        "cp /tmp/behave_tmpfile {}".format(file_path),
        "with sudo",
        machine_name=machine_name,
    )


@when("I delete the file `{file_path}`")
def when_i_delete_file(context, file_path):
    cmd = "rm -rf {}".format(file_path)
    cmd = 'sh -c "{}"'.format(cmd)
    when_i_run_command(context, cmd, "with sudo")


@when("I replace `{original}` in `{filename}` with `{new}`")
def when_i_replace_string_in_file(context, original, filename, new):
    new = new.replace("\\", r"\\")
    new = new.replace("/", r"\/")
    new = new.replace("&", r"\&")
    when_i_run_command(
        context,
        "sed -i 's/{original}/{new}/' {filename}".format(
            original=original, new=new, filename=filename
        ),
        "with sudo",
    )


@when("I replace `{original}` in `{filename}` with token `{token_name}`")
def when_i_replace_string_in_file_with_token(
    context, original, filename, token_name
):
    token = getattr(context.pro_config, token_name)
    when_i_replace_string_in_file(context, original, filename, token)


@then("I verify that files exist matching `{path_regex}`")
def there_should_be_files_matching_regex(context, path_regex):
    when_i_run_command(
        context, "ls {}".format(path_regex), "with sudo", verify_return=False
    )
    if context.process.returncode != 0:
        raise AssertionError("Missing expected files: {}".format(path_regex))


@then("I verify that no files exist matching `{path_regex}`")
def there_should_be_no_files_matching_regex(context, path_regex):
    when_i_run_command(
        context, "ls {}".format(path_regex), "with sudo", verify_return=False
    )
    if context.process.returncode == 0:
        raise AssertionError(
            "Unexpected files found: {}".format(context.process.stdout.strip())
        )


@when("I change config key `{key}` to use value `{yaml_value}`")
def change_config_key_to_use_value(context, key, yaml_value):
    yaml_value = process_template_vars(context, yaml_value)

    content = _get_file_contents(context, DEFAULT_CONFIG_FILE)
    cfg = yaml.safe_load(content)
    val = yaml.safe_load(yaml_value)
    cfg[key] = val
    new_content = yaml.dump(cfg)

    when_i_create_file_with_content(
        context, DEFAULT_CONFIG_FILE, text=new_content
    )


@when("I set `{key}` = `{json_value}` in json file `{filename}`")
def when_i_set_key_val_json_file(context, key, json_value, filename):
    json_value = process_template_vars(context, json_value)

    content_str = _get_file_contents(context, filename)
    content = json.loads(content_str)
    val = json.loads(json_value)
    content[key] = val
    new_content = json.dumps(content)

    when_i_create_file_with_content(context, filename, text=new_content)


@when("I move `{src_machine}` `{src_path}` to `{dest_machine}` `{dest_path}`")
def when_i_move_file_from_one_machine_to_another(
    context, src_machine, src_path, dest_machine, dest_path
):
    content = _get_file_contents(context, src_path, machine_name=src_machine)
    when_i_create_file_with_content(
        context, dest_path, machine_name=dest_machine, text=content
    )


@then(
    "I verify that `{file_path}` is owned by `{owner}` with permission `{permission}`"  # noqa
)
def then_i_verify_that_file_is_world_readable(
    context, file_path, owner, permission, machine_name=SUT
):
    when_i_run_command(
        context,
        "stat -c '%U:%G' {}".format(file_path),
        "as non-root",
        machine_name=machine_name,
    )
    assert_that(context.process.stdout.strip(), equal_to(owner))

    when_i_run_command(
        context,
        "stat -c '%a' {}".format(file_path),
        "as non-root",
        machine_name=machine_name,
    )

    assert_that(context.process.stdout.strip(), equal_to(permission))



import logging
from typing import Any, Dict, Optional, Tuple, Type

from uaclient import apt, event_logger, messages, system, util
from uaclient.entitlements import repo
from uaclient.entitlements.base import EntitlementWithMessage, UAEntitlement
from uaclient.types import (  # noqa: F401
    MessagingOperations,
    MessagingOperationsDict,
    StaticAffordance,
)

event = event_logger.get_event_logger()
LOG = logging.getLogger(util.replace_top_level_logger_name(__name__))

CPU_VENDOR_MAP = {"GenuineIntel": "intel"}


class RealtimeKernelEntitlement(repo.RepoEntitlement):
    name = "realtime-kernel"
    title = messages.REALTIME_TITLE
    description = messages.REALTIME_DESCRIPTION
    help_text = messages.REALTIME_HELP_TEXT
    help_doc_url = messages.urls.REALTIME_HOME_PAGE
    repo_key_file = "ubuntu-pro-realtime-kernel.gpg"
    apt_noninteractive = True
    supports_access_only = True
    supports_purge = False
    origin = "UbuntuRealtimeKernel"

    def _check_for_reboot(self) -> bool:
        """Check if system needs to be rebooted."""
        reboot_required = system.should_reboot(
            installed_pkgs=set(self.packages),
            installed_pkgs_regex=set(["linux-.*-realtime"]),
        )
        event.needs_reboot(reboot_required)
        return reboot_required

    def _get_variants(self) -> Dict[str, Type[UAEntitlement]]:
        return {
            GenericRealtime.variant_name: GenericRealtime,
            NvidiaTegraRealtime.variant_name: NvidiaTegraRealtime,
            RaspberryPiRealtime.variant_name: RaspberryPiRealtime,
            IntelIotgRealtime.variant_name: IntelIotgRealtime,
        }

    @property
    def default_variant(self):
        if self.access_only:
            return None
        return GenericRealtime

    @property
    def incompatible_services(self) -> Tuple[EntitlementWithMessage, ...]:
        from uaclient.entitlements.fips import (
            FIPSEntitlement,
            FIPSUpdatesEntitlement,
        )
        from uaclient.entitlements.livepatch import LivepatchEntitlement

        return (
            EntitlementWithMessage(
                FIPSEntitlement, messages.REALTIME_FIPS_INCOMPATIBLE
            ),
            EntitlementWithMessage(
                FIPSUpdatesEntitlement,
                messages.REALTIME_FIPS_UPDATES_INCOMPATIBLE,
            ),
            EntitlementWithMessage(
                LivepatchEntitlement, messages.REALTIME_LIVEPATCH_INCOMPATIBLE
            ),
        )

    @property
    def static_affordances(self) -> Tuple[StaticAffordance, ...]:
        return (
            (
                messages.REALTIME_ERROR_INSTALL_ON_CONTAINER,
                lambda: system.is_container(),
                False,
            ),
        )

    @property
    def messaging(
        self,
    ) -> MessagingOperationsDict:
        pre_enable = None  # type: Optional[MessagingOperations]
        if not self.access_only:
            pre_enable = [
                (
                    util.prompt_for_confirmation,
                    {
                        "msg": messages.REALTIME_PROMPT,
                        "default": True,
                    },
                )
            ]

        pre_disable = None  # type: Optional[MessagingOperations]
        if not self.purge:
            pre_disable = [
                (
                    util.prompt_for_confirmation,
                    {
                        "msg": messages.REALTIME_PRE_DISABLE_PROMPT,
                    },
                )
            ]

        return {
            "pre_enable": pre_enable,
            "pre_disable": pre_disable,
        }

    def remove_packages(self) -> None:
        packages = set(self.packages).intersection(
            set(apt.get_installed_packages_names())
        )
        if packages:
            apt.remove_packages(
                list(packages),
                messages.DISABLE_FAILED_TMPL.format(title=self.title),
            )


class RealtimeVariant(RealtimeKernelEntitlement):
    @property
    def incompatible_services(self) -> Tuple[EntitlementWithMessage, ...]:
        incompatible_variants = tuple(
            [
                EntitlementWithMessage(
                    cls,
                    messages.REALTIME_VARIANT_INCOMPATIBLE.format(
                        service=self.title, variant=cls.title
                    ),
                )
                for name, cls in self.other_variants.items()
            ]
        )
        return super().incompatible_services + incompatible_variants


class GenericRealtime(RealtimeVariant):
    variant_name = "generic"
    title = messages.REALTIME_GENERIC_TITLE
    description = messages.REALTIME_GENERIC_DESCRIPTION
    is_variant = True
    check_packages_are_installed = True

    @property
    def messaging(
        self,
    ) -> MessagingOperationsDict:
        messaging = super().messaging
        current_flavor = system.get_kernel_info().flavor
        if current_flavor != "generic":
            pre_enable = messaging.get("pre_enable") or []
            msg = messages.KERNEL_FLAVOR_CHANGE_WARNING_PROMPT.format(
                variant=self.variant_name,
                service=self.name,
                base_flavor="generic",
                current_flavor=current_flavor or "unknown",
            )
            pre_enable.append(
                (
                    util.prompt_for_confirmation,
                    {"msg": msg},
                )
            )
            messaging["pre_enable"] = pre_enable
        return messaging


class NvidiaTegraRealtime(RealtimeVariant):
    variant_name = "nvidia-tegra"
    title = messages.REALTIME_NVIDIA_TITLE
    description = messages.REALTIME_NVIDIA_DESCRIPTION
    is_variant = True
    check_packages_are_installed = True


class RaspberryPiRealtime(RealtimeVariant):
    variant_name = "raspi"
    title = messages.REALTIME_RASPI_TITLE
    description = messages.REALTIME_RASPI_DESCRIPTION
    is_variant = True
    check_packages_are_installed = True

    def variant_auto_select(self) -> bool:
        proc_file_path = "/proc/device-tree/model"
        try:
            model = system.load_file(proc_file_path).strip().lower()
            return "raspberry pi 4" in model or "raspberry pi 5" in model
        except Exception as e:
            LOG.info("Error while detecting if raspberry pi: %r", e)
            return False


class IntelIotgRealtime(RealtimeVariant):
    variant_name = "intel-iotg"
    title = messages.REALTIME_INTEL_TITLE
    description = messages.REALTIME_INTEL_DESCRIPTION
    is_variant = True
    check_packages_are_installed = True

    def verify_platform_checks(
        self, platform_checks: Dict[str, Any]
    ) -> Tuple[bool, Optional[messages.NamedMessage]]:
        vendor_id = system.get_cpu_info().cpuinfo_vendor_id or ""
        vendor_id = CPU_VENDOR_MAP.get(vendor_id, vendor_id)
        cpu_vendor_ids = platform_checks.get("cpu_vendor_ids", [])
        if vendor_id in cpu_vendor_ids:
            return True, None
        else:
            return False, messages.INAPPLICABLE_VENDOR_NAME.format(
                title=self.title,
                vendor=vendor_id,
                supported_vendors=",".join(cpu_vendor_ids),
            )



"""Tests related to uaclient.entitlement.base module."""

from functools import partial

import pytest

from uaclient import messages
from uaclient.entitlements.cis import CISEntitlement


@pytest.fixture
def cis_entitlement(entitlement_factory):
    return partial(
        entitlement_factory,
        CISEntitlement,
    )


class TestCISEntitlement:
    @pytest.mark.parametrize(
        "called_name,presentation_name,expected",
        (
            (
                "cis",
                "cis",
                {
                    "post_enable": [messages.CIS_POST_ENABLE],
                },
            ),
            (
                "cis",
                "usg",
                {
                    "pre_can_enable": [messages.CIS_IS_NOW_USG],
                    "post_enable": [messages.CIS_POST_ENABLE],
                },
            ),
            (
                "usg",
                "cis",
                {
                    "post_enable": [messages.CIS_USG_POST_ENABLE],
                },
            ),
            (
                "usg",
                "usg",
                {
                    "post_enable": [messages.CIS_USG_POST_ENABLE],
                },
            ),
        ),
    )
    def test_messages(
        self, called_name, presentation_name, expected, cis_entitlement
    ):
        entitlement = cis_entitlement(
            called_name=called_name,
            affordances={"presentedAs": presentation_name},
        )

        assert expected == entitlement.messaging

    @pytest.mark.parametrize(
        "called_name,expected",
        (
            ("cis", ["test-package"]),
            ("usg", ["usg"]),
        ),
    )
    def test_packages(self, called_name, expected, cis_entitlement):
        entitlement = cis_entitlement(
            called_name=called_name,
            directives={
                "additionalPackages": ["test-package"],
            },
        )

        assert expected == entitlement.packages

    @pytest.mark.parametrize(
        "called_name,expected",
        (
            ("cis", messages.CIS_TITLE),
            ("usg", messages.CIS_USG_TITLE),
        ),
    )
    def test_title(self, called_name, expected, cis_entitlement):
        entitlement = cis_entitlement(
            called_name=called_name,
        )

        assert expected == entitlement.title



import abc
import logging
from typing import Optional

# setup null handler for all API endpoints
logging.getLogger("ubuntupro").addHandler(logging.NullHandler())


class AbstractProgress(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def progress(
        self,
        *,
        total_steps: int,
        done_steps: int,
        previous_step_message: Optional[str],
        current_step_message: Optional[str]
    ):
        pass


class NullProgress(AbstractProgress):
    def progress(
        self,
        *,
        total_steps: int,
        done_steps: int,
        previous_step_message: Optional[str],
        current_step_message: Optional[str]
    ):
        pass


class ProgressWrapper:
    def __init__(self, progress_object: Optional[AbstractProgress] = None):
        if progress_object is not None:
            self.progress_object = progress_object
        else:
            self.progress_object = NullProgress()
        self.done_steps = 0
        self.total_steps = -1
        self.previous_step_message = None  # type: Optional[str]

    def progress(self, message: str):
        self.progress_object.progress(
            total_steps=self.total_steps,
            done_steps=self.done_steps,
            previous_step_message=self.previous_step_message,
            current_step_message=message,
        )
        self.previous_step_message = message
        self.done_steps += 1

    def finish(self):
        self.done_steps = self.total_steps
        self.progress_object.progress(
            total_steps=self.total_steps,
            done_steps=self.done_steps,
            previous_step_message=self.previous_step_message,
            current_step_message=None,
        )

    def emit(self, event: str, payload=None):
        """
        This is our secret event system. We use it internally to insert prompts
        and extra messages in the middle of operations at certain points.
        We don't consider this stable enough to expose to the public API.
        """
        if hasattr(self.progress_object, "_on_event"):
            self.progress_object._on_event(event, payload)

    def is_interactive(self) -> bool:
        if hasattr(self.progress_object, "is_interactive"):
            return self.progress_object.is_interactive
        else:
            return False



import os

import mock
import pytest

from uaclient import lock
from uaclient.defaults import DEFAULT_DATA_DIR
from uaclient.exceptions import InvalidLockFile, LockHeldError
from uaclient.files.notices import Notice
from uaclient.lock import RetryLock
from uaclient.messages import E_INVALID_LOCK_FILE, LOCK_HELD

M_PATH = "uaclient.lock."
M_PATH_UACONFIG = "uaclient.config.UAConfig."


class TestRetryLock:
    @mock.patch("uaclient.lock.check_lock_info", return_value=(-1, ""))
    @mock.patch("os.getpid", return_value=123)
    @mock.patch("uaclient.files.notices.NoticesManager.add")
    def test_creates_and_releases_lock(
        self,
        m_add_notice,
        _m_getpid,
        _m_check_lock_info,
    ):
        def test_function():
            return mock.sentinel.success

        with mock.patch.object(lock, "lock_data_file") as m_lock_file:
            with lock.RetryLock(lock_holder="some operation"):
                ret = test_function()

        assert mock.sentinel.success == ret
        assert [
            mock.call(
                lock.LockData(lock_pid="123", lock_holder="some operation")
            )
        ] == m_lock_file.write.call_args_list
        lock_msg = "Operation in progress: some operation"
        assert [
            mock.call(Notice.OPERATION_IN_PROGRESS, lock_msg)
        ] == m_add_notice.call_args_list
        assert 1 == m_lock_file.delete.call_count

    @mock.patch("uaclient.lock.check_lock_info", return_value=(-1, ""))
    @mock.patch("os.getpid", return_value=123)
    @mock.patch("uaclient.files.notices.NoticesManager.add")
    def test_creates_and_releases_lock_when_error_occurs(
        self,
        m_add_notice,
        _m_getpid,
        _m_check_lock_info,
    ):
        def test_function():
            raise RuntimeError("test")

        with pytest.raises(RuntimeError) as exc:
            with mock.patch.object(lock, "lock_data_file") as m_lock_file:
                with RetryLock(lock_holder="some operation"):
                    test_function()

        assert "test" == str(exc.value)
        assert [
            mock.call(
                lock.LockData(lock_pid="123", lock_holder="some operation")
            )
        ] == m_lock_file.write.call_args_list
        lock_msg = "Operation in progress: some operation"
        assert [
            mock.call(Notice.OPERATION_IN_PROGRESS, lock_msg)
        ] == m_add_notice.call_args_list
        assert 1 == m_lock_file.delete.call_count

    @mock.patch(M_PATH + "time.sleep")
    @mock.patch(
        M_PATH + "RetryLock.grab_lock",
        side_effect=[
            LockHeldError(
                lock_request="request", lock_holder="holder", pid=10
            ),
            LockHeldError(
                lock_request="request", lock_holder="holder", pid=10
            ),
            None,
        ],
    )
    def test_spins_when_lock_held(self, m_single_attempt_lock_enter, m_sleep):
        with lock.RetryLock(
            lock_holder="request", sleep_time=1, max_retries=3
        ):
            pass

        assert [
            mock.call(),
            mock.call(),
            mock.call(),
        ] == m_single_attempt_lock_enter.call_args_list
        assert [mock.call(1), mock.call(1)] == m_sleep.call_args_list

    @mock.patch(M_PATH + "time.sleep")
    @mock.patch(
        M_PATH + "RetryLock.grab_lock",
        side_effect=[
            LockHeldError(
                lock_request="request", lock_holder="holder", pid=10
            ),
            LockHeldError(
                lock_request="request", lock_holder="holder", pid=10
            ),
            None,
        ],
    )
    def test_raises_lock_held_after_max_retries(
        self, m_single_attempt_lock_enter, m_sleep
    ):
        with pytest.raises(LockHeldError) as exc:
            with RetryLock(lock_holder="request", sleep_time=1, max_retries=2):
                pass

        assert (
            "Unable to perform: request.\n"
            + LOCK_HELD.format(lock_holder="holder", pid=10)
            == exc.value.msg
        )

        assert [
            mock.call(),
            mock.call(),
        ] == m_single_attempt_lock_enter.call_args_list
        assert [mock.call(1)] == m_sleep.call_args_list


class TestCheckLockInfo:
    @pytest.mark.parametrize("lock_content", ((""), ("corrupted")))
    @mock.patch("uaclient.system.load_file")
    def test_raise_exception_for_corrupted_lock(
        self,
        m_load_file,
        lock_content,
    ):
        m_load_file.return_value = lock_content

        expected_msg = E_INVALID_LOCK_FILE.format(
            lock_file_path=os.path.join(DEFAULT_DATA_DIR, "lock")
        )

        with pytest.raises(InvalidLockFile) as exc_info:
            lock.check_lock_info()

        assert expected_msg.msg == exc_info.value.msg
        assert m_load_file.call_count == 1



import enum
from typing import Optional

from uaclient import messages


@enum.unique
class ApplicationStatus(enum.Enum):
    """
    An enum to represent the current application status of an entitlement
    """

    ENABLED = object()
    DISABLED = object()
    WARNING = object()


@enum.unique
class ContractStatus(enum.Enum):
    """
    An enum to represent whether a user is entitled to an entitlement

    (The value of each member is the string that will be used in status
    output.)
    """

    ENTITLED = "yes"
    UNENTITLED = "no"


@enum.unique
class ApplicabilityStatus(enum.Enum):
    """
    An enum to represent whether an entitlement could apply to this machine
    """

    APPLICABLE = object()
    INAPPLICABLE = object()


@enum.unique
class UserFacingAvailability(enum.Enum):
    """
    An enum representing whether a service could be available for a machine.

    'Availability' means whether a service is available to machines with this
    architecture, series and kernel. Whether a contract is entitled to use
    the specific service is determined by the contract level.

    This enum should only be used in display code, it should not be used in
    business logic.
    """

    AVAILABLE = "yes"
    UNAVAILABLE = "no"


@enum.unique
class UserFacingConfigStatus(enum.Enum):
    """
    An enum representing the user-visible config status of Pro system.

    This enum will be used in display code and will be written to status.json
    """

    INACTIVE = "inactive"  # No Pro config commands/daemons
    ACTIVE = "active"  # Pro command is running
    REBOOTREQUIRED = "reboot-required"  # System Reboot required


@enum.unique
class UserFacingStatus(enum.Enum):
    """
    An enum representing the states we will display in status output.

    This enum should only be used in display code, it should not be used in
    business logic.
    """

    ACTIVE = "enabled"
    INACTIVE = "disabled"
    INAPPLICABLE = "n/a"
    UNAVAILABLE = "—"
    WARNING = "warning"


@enum.unique
class CanEnableFailureReason(enum.Enum):
    """
    An enum representing the reasons an entitlement can't be enabled.
    """

    NOT_ENTITLED = object()
    ALREADY_ENABLED = object()
    INAPPLICABLE = object()
    IS_BETA = object()
    INCOMPATIBLE_SERVICE = object()
    INACTIVE_REQUIRED_SERVICES = object()
    ACCESS_ONLY_NOT_SUPPORTED = object()
    ONLY_ACCESS_ONLY_SUPPORTED = object()


class CanEnableFailure:
    def __init__(
        self,
        reason: CanEnableFailureReason,
        message: Optional[messages.NamedMessage] = None,
    ) -> None:
        self.reason = reason
        self.message = message


@enum.unique
class CanDisableFailureReason(enum.Enum):
    """
    An enum representing the reasons an entitlement can't be disabled.
    """

    ALREADY_DISABLED = object()
    NOT_APPLICABLE = object()
    ACTIVE_DEPENDENT_SERVICES = object()
    PURGE_NOT_SUPPORTED = object()
    NOT_FOUND_DEPENDENT_SERVICE = object()
    NO_PURGE_WITHOUT_ORIGIN = object()


class CanDisableFailure:
    def __init__(
        self,
        reason: CanDisableFailureReason,
        message: Optional[messages.NamedMessage] = None,
    ) -> None:
        self.reason = reason
        self.message = message

    @property
    def message_value(self) -> str:
        return self.message.msg if self.message else ""



from uaclient import event_logger, messages
from uaclient.api.u.pro.security.status.reboot_required.v1 import (
    _reboot_required,
)
from uaclient.cli.commands import ProCommand
from uaclient.cli.parser import HelpCategory

event = event_logger.get_event_logger()


def action_reboot_required(args, *, cfg, **kwargs):
    result = _reboot_required(cfg)
    event.info(result.reboot_required)
    return 0


def action_system(args, *, cfg, **kwargs):
    # Avoiding a circular import
    from uaclient.cli import get_parser

    get_parser().print_help_for_command("system")


reboot_required_subcommand = ProCommand(
    "reboot-required",
    help=messages.CLI_SYSTEM_REBOOT_REQUIRED,
    description=messages.CLI_SYSTEM_REBOOT_REQUIRED_DESC,
    action=action_reboot_required,
    preserve_description=True,
)

system_command = ProCommand(
    "system",
    help=messages.CLI_ROOT_SYSTEM,
    description=messages.CLI_SYSTEM_DESC,
    action=action_system,
    help_category=HelpCategory.QUICKSTART,
    help_position=4,
    subcommands=[reboot_required_subcommand],
)



import os

import mock
import pytest

from uaclient import apt, exceptions, http, messages
from uaclient.config import (
    UA_CONFIGURABLE_KEYS,
    VALID_UA_CONFIG_KEYS,
    get_config_path,
    parse_config,
)
from uaclient.conftest import FakeNotice
from uaclient.defaults import DEFAULT_CONFIG_FILE
from uaclient.files import notices, user_config_file
from uaclient.files.notices import NoticesManager
from uaclient.util import depth_first_merge_overlay_dict
from uaclient.yaml import safe_dump

KNOWN_DATA_PATHS = (("machine-id", "machine-id"),)
M_PATH = "uaclient.entitlements."


class TestNotices:
    @pytest.mark.parametrize(
        "notices,expected",
        (
            ([], ()),
            (
                [[FakeNotice.reboot_script_failed, "a1"]],
                ["a1"],
            ),
            (
                [
                    [FakeNotice.reboot_required, "a1"],
                    [FakeNotice.reboot_script_failed, "a2"],
                ],
                [
                    "a1",
                    "a2",
                ],
            ),
            (
                [
                    [FakeNotice.reboot_required, "a1"],
                    [FakeNotice.reboot_required, "a1"],
                ],
                [
                    "a1",
                ],
            ),
        ),
    )
    def test_add_notice_avoids_duplicates(
        self,
        notices,
        expected,
    ):
        notice = NoticesManager()
        assert [] == notice.list()
        for notice_ in notices:
            notice.add(*notice_)
        if notices:
            assert expected == notice.list()
        else:
            assert [] == notice.list()

    @pytest.mark.parametrize(
        "_notices",
        (
            ([]),
            ([[FakeNotice.reboot_required]]),
            (
                [
                    [FakeNotice.reboot_required],
                    [FakeNotice.reboot_script_failed],
                ]
            ),
        ),
    )
    @mock.patch("uaclient.util.we_are_currently_root", return_value=False)
    def test_add_notice_fails_as_nonroot(
        self,
        m_we_are_currently_root,
        _notices,
    ):
        assert [] == notices.list()
        for notice_ in _notices:
            notices.add(*notice_)
        assert [] == notices.list()

    @pytest.mark.parametrize(
        "notices_,removes,expected",
        (
            ([], [FakeNotice.reboot_required], []),
            (
                [[FakeNotice.reboot_script_failed]],
                [FakeNotice.reboot_script_failed],
                [],
            ),
            (
                [
                    [FakeNotice.reboot_required],
                    [FakeNotice.reboot_script_failed],
                ],
                [FakeNotice.reboot_required],
                ["notice_a2"],
            ),
            (
                [
                    [FakeNotice.reboot_required],
                    [FakeNotice.reboot_script_failed],
                    [FakeNotice.enable_reboot_required],
                ],
                [
                    FakeNotice.reboot_required,
                    FakeNotice.reboot_script_failed,
                ],
                ["notice_b"],
            ),
        ),
    )
    def test_remove_notice_removes_matching(
        self,
        notices_,
        removes,
        expected,
    ):
        for notice_ in notices_:
            notices.add(*notice_)
        for label in removes:
            notices.remove(label)
        assert expected == notices.list()


CFG_BASE_CONTENT = """\
# Ubuntu Pro client config file.
# If you modify this file, run "pro refresh config" to ensure changes are
# picked up by Ubuntu Pro client.

contract_url: https://contracts.canonical.com
data_dir: /var/lib/ubuntu-advantage
log_file: /var/log/ubuntu-advantage.log
log_level: debug
security_url: https://ubuntu.com/security
"""

CFG_FEATURES_CONTENT = """\
# Ubuntu Pro client config file.
# If you modify this file, run "pro refresh config" to ensure changes are
# picked up by Ubuntu Pro client.

contract_url: https://contracts.canonical.com
data_dir: /var/lib/ubuntu-advantage
features:
  extra_security_params:
    hide: true
  new: 2
  show_beta: true
log_file: /var/log/ubuntu-advantage.log
log_level: debug
security_url: https://ubuntu.com/security
settings_overrides:
  c: 1
  d: 2
"""

USER_CFG_DICT = {
    "apt_http_proxy": None,
    "apt_https_proxy": None,
    "apt_news": True,
    "apt_news_url": "https://motd.ubuntu.com/aptnews.json",
    "global_apt_http_proxy": None,
    "global_apt_https_proxy": None,
    "ua_apt_http_proxy": None,
    "ua_apt_https_proxy": None,
    "http_proxy": None,
    "https_proxy": None,
    "update_messaging_timer": 21600,
    "metering_timer": 14400,
    "vulnerability_data_url_prefix": "https://security-metadata.canonical.com/oval/",  # noqa
    "lxd_guest_attach": user_config_file.LXDGuestAttachEnum.OFF,
}


class TestUserConfigKeys:
    @pytest.mark.parametrize("attr_name", UA_CONFIGURABLE_KEYS)
    @mock.patch("uaclient.config.user_config_file.user_config.write")
    def test_user_configurable_keys_set_user_config(
        self, write, attr_name, tmpdir, FakeConfig
    ):
        """Getters and settings are available fo UA_CONFIGURABLE_KEYS."""
        cfg = FakeConfig()
        assert USER_CFG_DICT[attr_name] == getattr(cfg, attr_name, None)
        cfg_non_members = ("apt_http_proxy", "apt_https_proxy")
        if attr_name not in cfg_non_members:
            setattr(cfg, attr_name, attr_name + "value")
            assert attr_name + "value" == getattr(cfg, attr_name)
            assert attr_name + "value" == getattr(cfg.user_config, attr_name)


class TestProcessConfig:
    @pytest.mark.parametrize(
        [
            "http_proxy",
            "https_proxy",
            "snap_is_snapd_installed",
            "snap_http_val",
            "snap_https_val",
            "livepatch_enabled",
            "livepatch_http_val",
            "livepatch_https_val",
            "snap_livepatch_msg",
            "global_https",
            "global_http",
            "ua_https",
            "ua_http",
            "apt_https",
            "apt_http",
            "lxd_guest_attach",
            "is_attached",
        ],
        [
            (
                "http",
                "https",
                False,
                None,
                None,
                False,
                None,
                None,
                "",
                None,
                None,
                None,
                None,
                None,
                None,
                user_config_file.LXDGuestAttachEnum.OFF,
                True,
            ),
            (
                "http",
                "https",
                True,
                None,
                None,
                False,
                None,
                None,
                "",
                None,
                None,
                None,
                None,
                "apt_https",
                "apt_http",
                user_config_file.LXDGuestAttachEnum.OFF,
                True,
            ),
            (
                "http",
                "https",
                False,
                None,
                None,
                True,
                None,
                None,
                "",
                "global_https",
                "global_http",
                None,
                None,
                None,
                None,
                user_config_file.LXDGuestAttachEnum.OFF,
                True,
            ),
            (
                "http",
                "https",
                True,
                None,
                None,
                True,
                None,
                None,
                "",
                None,
                None,
                "ua_https",
                "ua_http",
                None,
                None,
                user_config_file.LXDGuestAttachEnum.OFF,
                True,
            ),
            (
                None,
                None,
                True,
                None,
                None,
                True,
                None,
                None,
                "",
                "global_https",
                "global_http",
                None,
                None,
                "apt_https",
                "apt_http",
                user_config_file.LXDGuestAttachEnum.OFF,
                True,
            ),
            (
                None,
                None,
                True,
                "one",
                None,
                True,
                None,
                None,
                "snap",
                "global_https",
                "global_http",
                "ua_https",
                "ua_http",
                "apt_https",
                "apt_http",
                user_config_file.LXDGuestAttachEnum.OFF,
                True,
            ),
            (
                None,
                None,
                True,
                "one",
                "two",
                True,
                None,
                None,
                "snap",
                None,
                "global_http",
                None,
                None,
                None,
                "apt_http",
                user_config_file.LXDGuestAttachEnum.OFF,
                True,
            ),
            (
                None,
                None,
                True,
                "one",
                "two",
                True,
                "three",
                None,
                "snap, livepatch",
                "global_htttps",
                None,
                "ua_https",
                None,
                "apt_https",
                None,
                user_config_file.LXDGuestAttachEnum.OFF,
                True,
            ),
            (
                None,
                None,
                True,
                "one",
                "two",
                True,
                "three",
                "four",
                "snap, livepatch",
                "global_https",
                None,
                None,
                "ua_http",
                None,
                None,
                user_config_file.LXDGuestAttachEnum.OFF,
                True,
            ),
            (
                None,
                None,
                False,
                None,
                None,
                True,
                "three",
                "four",
                "livepatch",
                None,
                None,
                None,
                None,
                None,
                None,
                user_config_file.LXDGuestAttachEnum.OFF,
                False,
            ),
        ],
    )
    @mock.patch("uaclient.api.u.pro.status.is_attached.v1._is_attached")
    @mock.patch("uaclient.config.state_files.lxd_pro_config_file.write")
    @mock.patch(
        "uaclient.api.u.pro.status.enabled_services.v1._enabled_services"
    )
    @mock.patch("uaclient.http.validate_proxy")
    @mock.patch("uaclient.livepatch.get_config_option_value")
    @mock.patch("uaclient.livepatch.configure_livepatch_proxy")
    @mock.patch("uaclient.snap.get_config_option_value")
    @mock.patch("uaclient.snap.configure_snap_proxy")
    @mock.patch("uaclient.snap.is_snapd_installed")
    @mock.patch("uaclient.apt.setup_apt_proxy")
    @mock.patch("uaclient.config.user_config_file.user_config.write")
    def test_process_config(
        self,
        m_write,
        m_apt_configure_proxy,
        m_snap_is_snapd_installed,
        m_snap_configure_proxy,
        m_snap_get_config_option,
        m_livepatch_configure_proxy,
        m_livepatch_get_config_option,
        m_validate_proxy,
        m_enabled_services,
        m_lxd_pro_config_file_write,
        m_is_attached,
        http_proxy,
        https_proxy,
        snap_is_snapd_installed,
        snap_http_val,
        snap_https_val,
        livepatch_enabled,
        livepatch_http_val,
        livepatch_https_val,
        snap_livepatch_msg,
        global_https,
        global_http,
        ua_https,
        ua_http,
        apt_https,
        apt_http,
        lxd_guest_attach,
        is_attached,
        capsys,
        tmpdir,
        FakeConfig,
    ):
        m_snap_is_snapd_installed.return_value = snap_is_snapd_installed
        m_snap_get_config_option.side_effect = [snap_http_val, snap_https_val]
        m_is_attached.return_value = is_attached

        _m_livepatch = mock.MagicMock()
        type(_m_livepatch).name = mock.PropertyMock(return_value="livepatch")
        m_enabled_services.return_value = mock.MagicMock(
            enabled_services=[_m_livepatch]
        )

        m_livepatch_get_config_option.side_effect = [
            livepatch_http_val,
            livepatch_https_val,
        ]
        cfg = FakeConfig({"data_dir": tmpdir.strpath})
        cfg.user_config.apt_http_proxy = apt_http
        cfg.user_config.apt_https_proxy = apt_https
        cfg.user_config.global_apt_https_proxy = global_https
        cfg.user_config.global_apt_http_proxy = global_http
        cfg.user_config.ua_apt_https_proxy = ua_https
        cfg.user_config.ua_apt_http_proxy = ua_http
        cfg.user_config.http_proxy = http_proxy
        cfg.user_config.https_proxy = https_proxy
        cfg.user_config.update_messaging_timer = 21600
        cfg.user_config.metering_timer = 0
        cfg.user_config.lxd_guest_attach = lxd_guest_attach

        if global_https is None and apt_https is not None:
            global_https = apt_https
        if global_http is None and apt_http is not None:
            global_http = apt_http

        exc = False
        if global_https or global_http:
            if ua_https or ua_http:
                exc = True
                with pytest.raises(
                    exceptions.UbuntuProError,
                    match=messages.E_INVALID_PROXY_COMBINATION.msg,
                ):
                    cfg.process_config()
        if exc is False:
            cfg.process_config()

            assert [
                mock.call(
                    "http", global_http, http.PROXY_VALIDATION_APT_HTTP_URL
                ),
                mock.call(
                    "https", global_https, http.PROXY_VALIDATION_APT_HTTPS_URL
                ),
                mock.call("http", ua_http, http.PROXY_VALIDATION_APT_HTTP_URL),
                mock.call(
                    "https", ua_https, http.PROXY_VALIDATION_APT_HTTPS_URL
                ),
                mock.call(
                    "http", http_proxy, http.PROXY_VALIDATION_SNAP_HTTP_URL
                ),
                mock.call(
                    "https", https_proxy, http.PROXY_VALIDATION_SNAP_HTTPS_URL
                ),
            ] == m_validate_proxy.call_args_list

            if global_http or global_https:
                assert [
                    mock.call(
                        global_http, global_https, apt.AptProxyScope.GLOBAL
                    )
                ] == m_apt_configure_proxy.call_args_list
            elif ua_http or ua_https:
                assert [
                    mock.call(ua_http, ua_https, apt.AptProxyScope.UACLIENT)
                ] == m_apt_configure_proxy.call_args_list
            else:
                assert [] == m_apt_configure_proxy.call_args_list

            if snap_is_snapd_installed:
                assert [
                    mock.call(http_proxy, https_proxy)
                ] == m_snap_configure_proxy.call_args_list

            if livepatch_enabled:
                assert [
                    mock.call(http_proxy, https_proxy)
                ] == m_livepatch_configure_proxy.call_args_list

            expected_out = ""
            if snap_livepatch_msg:
                expected_out = messages.PROXY_DETECTED_BUT_NOT_CONFIGURED.format(  # noqa: E501
                    services=snap_livepatch_msg
                )

            out, err = capsys.readouterr()
            expected_out = """
                Using deprecated "{apt}" config field.
                Please migrate to using "{global_}"
            """
            if apt_http and not global_http:
                assert (
                    expected_out.format(
                        apt=apt_http, global_=global_http
                    ).strip()
                    == out.strip()
                )
            if apt_https and not global_https:
                assert (
                    expected_out.format(
                        apt=apt_https, global_=global_https
                    ).strip()
                    == out.strip()
                )
            assert "" == err
            if (
                lxd_guest_attach != user_config_file.LXDGuestAttachEnum.OFF
                or is_attached
            ):
                assert m_lxd_pro_config_file_write.call_args_list == [
                    mock.call(mock.ANY)
                ]

    def test_process_config_errors_for_wrong_timers(self, FakeConfig):
        cfg = FakeConfig()
        cfg.user_config.update_messaging_timer = "wrong"

        with pytest.raises(
            exceptions.UbuntuProError,
            match=(
                "Cannot set update_messaging_timer to wrong: <value> for "
                "interval must be a positive integer."
            ),
        ):
            cfg.process_config()


class TestParseConfig:
    @mock.patch("uaclient.config.os.path.exists", return_value=False)
    @mock.patch("uaclient.contract.get_available_resources")
    def test_parse_config_uses_defaults_when_no_config_present(
        self, _m_resources, m_exists
    ):
        with mock.patch.dict("uaclient.config.os.environ", values={}):
            config, _ = parse_config()
        expected_calls = [
            mock.call("/etc/ubuntu-advantage/uaclient.conf"),
        ]
        assert expected_calls == m_exists.call_args_list
        expected_default_config = {
            "contract_url": "https://contracts.canonical.com",
            "security_url": "https://ubuntu.com/security",
            "data_dir": "/var/lib/ubuntu-advantage",
            "log_file": "/var/log/ubuntu-advantage.log",
            "log_level": "debug",
        }
        assert expected_default_config == config

    @pytest.mark.parametrize(
        "config_dict,expected_invalid_keys",
        (
            ({"contract_url": "http://abc", "security_url": "http:xyz"}, []),
            (
                {"contract_urs": "http://abc", "security_url": "http:xyz"},
                ["contract_urs"],
            ),
        ),
    )
    def test_parse_config_returns_invalid_keys(
        self, config_dict, expected_invalid_keys, tmpdir
    ):
        config_file = tmpdir.join("uaclient.conf")
        config_file.write(safe_dump(config_dict))
        env_vars = {"UA_CONFIG_FILE": config_file.strpath}
        with mock.patch.dict("uaclient.config.os.environ", values=env_vars):
            cfg, invalid_keys = parse_config(config_file.strpath)
        assert set(expected_invalid_keys) == invalid_keys
        for key, value in config_dict.items():
            if key in VALID_UA_CONFIG_KEYS:
                assert config_dict[key] == cfg[key]

    @pytest.mark.parametrize(
        "envvar_name,envvar_val,field,expected_val",
        [
            # not on allowlist
            (
                "UA_CONTRACT_URL",
                "https://contract",
                "contract_url",
                "https://contracts.canonical.com",
            ),
            # on allowlist
            (
                "UA_security_URL",
                "https://security",
                "security_url",
                "https://security",
            ),
            (
                "ua_data_dir",
                "~/somedir",
                "data_dir",
                "{}/somedir".format(os.path.expanduser("~")),
            ),
            ("Ua_LoG_FiLe", "some.log", "log_file", "some.log"),
            ("UA_LOG_LEVEL", "debug", "log_level", "debug"),
        ],
    )
    @mock.patch("uaclient.config.os.path.exists", return_value=False)
    @mock.patch("uaclient.contract.get_available_resources")
    def test_parse_config_scrubs_user_environ_values(
        self,
        _m_resources,
        m_exists,
        envvar_name,
        envvar_val,
        field,
        expected_val,
    ):
        user_values = {envvar_name: envvar_val}
        with mock.patch.dict("uaclient.config.os.environ", values=user_values):
            config, _ = parse_config()
        assert expected_val == config[field]

    @mock.patch("uaclient.config.os.path.exists", return_value=False)
    def test_parse_config_scrubs_user_environ_values_features(self, m_exists):
        user_values = {
            "UA_FEATURES_X_Y_Z": "XYZ_VAL",
            "UA_FEATURES_A_B_C": "ABC_VAL",
        }
        with mock.patch.dict("uaclient.config.os.environ", values=user_values):
            config, _ = parse_config()
        expected_config = {
            "features": {"a_b_c": "ABC_VAL", "x_y_z": "XYZ_VAL"}
        }
        assert expected_config["features"] == config["features"]

    @pytest.mark.parametrize(
        "env_var,env_value", [("UA_SECURITY_URL", "ht://security")]
    )
    @mock.patch("uaclient.config.os.path.exists", return_value=False)
    def test_parse_raises_errors_on_invalid_urls(
        self, _m_exists, env_var, env_value
    ):
        user_values = {env_var: env_value}  # no acceptable url scheme
        with mock.patch.dict("uaclient.config.os.environ", values=user_values):
            with pytest.raises(exceptions.UbuntuProError) as excinfo:
                parse_config()
        expected_msg = "Invalid url in config. {}: {}".format(
            env_var.replace("UA_", "").lower(), env_value
        )
        assert expected_msg == excinfo.value.msg

    @mock.patch("uaclient.config.os.path.exists")
    @mock.patch("uaclient.system.load_file")
    def test_parse_reads_yaml_from_environ_values(
        self, m_load_file, m_path_exists
    ):
        m_load_file.return_value = "test: true\nfoo: bar"
        m_path_exists.side_effect = [False, True]

        user_values = {"UA_FEATURES_TEST": "test.yaml"}
        with mock.patch.dict("uaclient.config.os.environ", values=user_values):
            cfg, _ = parse_config()

        assert {"test": True, "foo": "bar"} == cfg["features"]["test"]

    @mock.patch("uaclient.config.os.path.exists")
    def test_parse_raise_exception_when_environ_yaml_file_does_not_exist(
        self, m_path_exists
    ):
        m_path_exists.return_value = False
        user_values = {"UA_FEATURES_TEST": "test.yaml"}
        with mock.patch.dict("uaclient.config.os.environ", values=user_values):
            with pytest.raises(exceptions.UbuntuProError) as excinfo:
                parse_config()

        expected_msg = "Could not find yaml file: test.yaml"
        assert expected_msg == excinfo.value.msg.strip()


class TestFeatures:
    @pytest.mark.parametrize(
        "cfg_features,expected, warnings",
        (
            ({}, {}, None),
            (None, {}, None),
            (
                "badstring",
                {},
                "Unexpected uaclient.conf features value."
                " Expected dict, but found %s",
            ),
            ({"feature1": "value1"}, {"feature1": "value1"}, None),
            (
                {"feature1": "value1", "feature2": False},
                {"feature1": "value1", "feature2": False},
                None,
            ),
        ),
    )
    @mock.patch("uaclient.config.LOG.warning")
    def test_features_are_a_property_of_uaconfig(
        self,
        m_log_warning,
        cfg_features,
        expected,
        warnings,
        FakeConfig,
    ):
        user_cfg = {"features": cfg_features}
        cfg = FakeConfig(cfg_overrides=user_cfg)
        assert expected == cfg.features
        if warnings:
            assert [
                mock.call(warnings, cfg_features)
            ] == m_log_warning.call_args_list


class TestDepthFirstMergeOverlayDict:
    @pytest.mark.parametrize(
        "base_dict, overlay_dict, expected_dict",
        [
            ({"a": 1, "b": 2}, {"c": 3}, {"a": 1, "b": 2, "c": 3}),
            (
                {"a": 1, "b": {"c": 2, "d": 3}},
                {"a": 1, "b": {"c": 10}},
                {"a": 1, "b": {"c": 10, "d": 3}},
            ),
            (
                {"a": 1, "b": {"c": 2, "d": 3}},
                {"d": {"f": 20}},
                {"a": 1, "b": {"c": 2, "d": 3}, "d": {"f": 20}},
            ),
            ({"a": 1, "b": 2}, {}, {"a": 1, "b": 2}),
            ({"a": 1, "b": 2}, {"a": "test"}, {"a": "test", "b": 2}),
            ({}, {"a": 1, "b": 2}, {"a": 1, "b": 2}),
            ({"a": []}, {"a": [1, 2, 3]}, {"a": [1, 2, 3]}),
            ({"a": [5, 6]}, {"a": [1, 2, 3]}, {"a": [1, 2, 3]}),
            ({"a": [{"b": 1}]}, {"a": [{"c": 2}]}, {"a": [{"b": 1, "c": 2}]}),
        ],
    )
    def test_depth_first_merge_dict(
        self, base_dict, overlay_dict, expected_dict
    ):
        depth_first_merge_overlay_dict(base_dict, overlay_dict)
        assert expected_dict == base_dict


class TestGetConfigPath:
    def test_get_config_path_from_env_var(self):
        with mock.patch.dict(
            "uaclient.config.os.environ", values={"UA_CONFIG_FILE": "test"}
        ):
            assert "test" == get_config_path()

    def test_get_default_config_path(self):
        with mock.patch.dict("uaclient.config.os.environ", values={}):
            assert DEFAULT_CONFIG_FILE == get_config_path()


class TestConfigShow:
    @mock.patch("uaclient.config.user_config_file.user_config.write")
    def test_redact_config_data(self, _write, FakeConfig):
        cfg = FakeConfig()

        setattr(
            cfg.user_config,
            "apt_http_proxy",
            "http://username:password@proxy:port",
        )
        setattr(
            cfg.user_config,
            "apt_https_proxy",
            "http://username:password@proxy:port",
        )
        setattr(
            cfg.user_config,
            "global_apt_http_proxy",
            "http://username:password@proxy:port",
        )
        setattr(
            cfg.user_config,
            "global_apt_https_proxy",
            "http://username:password@proxy:port",
        )
        setattr(
            cfg.user_config,
            "ua_apt_http_proxy",
            "http://username:password@proxy:port",
        )
        setattr(
            cfg.user_config,
            "ua_apt_https_proxy",
            "http://username:password@proxy:port",
        )
        setattr(
            cfg.user_config,
            "http_proxy",
            "http://username:password@proxy:port",
        )
        setattr(cfg.user_config, "https_proxy", "https://www.example.com")

        user_config_file_object = user_config_file.UserConfigFileObject()
        redacted_config = user_config_file_object.redact_config_data(
            cfg.user_config
        )

        # Assert that proxy configurations are redacted
        assert getattr(redacted_config, "apt_http_proxy") == "<REDACTED>"
        assert getattr(redacted_config, "apt_https_proxy") == "<REDACTED>"
        assert (
            getattr(redacted_config, "global_apt_http_proxy") == "<REDACTED>"
        )
        assert (
            getattr(redacted_config, "global_apt_https_proxy") == "<REDACTED>"
        )
        assert getattr(redacted_config, "ua_apt_http_proxy") == "<REDACTED>"
        assert getattr(redacted_config, "ua_apt_https_proxy") == "<REDACTED>"
        assert getattr(redacted_config, "http_proxy") == "<REDACTED>"
        assert (
            getattr(redacted_config, "https_proxy")
            == "https://www.example.com"
        )

        # Assert that redacting multiple times does not change the result
        redacted_again = user_config_file_object.redact_config_data(
            redacted_config
        )
        assert redacted_config == redacted_again



import pytest

from uaclient.files.state_files import _services_once_enable_preprocess_data


class TestServicesOnceEnabledPreprocessData:
    @pytest.mark.parametrize(
        "content",
        (
            ({"fips-updates": True}),
            ({"fips_updates": True}),
        ),
    )
    def test_add_service(self, content):
        assert _services_once_enable_preprocess_data(content) == {
            "fips_updates": True
        }



import os
import tempfile

from apport.hookutils import attach_file_if_exists
from uaclient import defaults
from uaclient.actions import APPARMOR_PROFILES, collect_logs
from uaclient.config import UAConfig
from uaclient.files.state_files import timer_jobs_state_file


def add_info(report, ui=None):
    report["LaunchpadPrivate"] = "1"
    report["LaunchpadSubscribe"] = "ua-client"
    cfg = UAConfig()
    apparmor_files = [os.path.basename(f) for f in APPARMOR_PROFILES]
    with tempfile.TemporaryDirectory() as output_dir:
        collect_logs(cfg, output_dir)
        auto_include_log_files = {
            "cloud-id.txt",
            "cloud-id.txt-error",
            "pro-status.json",
            "pro-status.json-error",
            "livepatch-status.txt",
            "livepatch-status.txt-error",
            "pro-journal.txt",
            "apparmor_logs.txt",
            *apparmor_files,
            os.path.basename(cfg.cfg_path),
            os.path.basename(cfg.log_file),
            os.path.basename(timer_jobs_state_file.path),
            os.path.basename(defaults.CONFIG_DEFAULTS["log_file"]),
        }
        for f in auto_include_log_files:
            attach_file_if_exists(report, os.path.join(output_dir, f), key=f)



from typing import List

from uaclient.api.api import APIEndpoint
from uaclient.api.data_types import AdditionalInfo
from uaclient.api.u.pro.security.fix._common import get_expected_overall_status

# Some of these imports are intentionally not used in this module.
# The rationale is that we want users to import such Data Objects
# directly from the associated endpoints and not through the _common module
from uaclient.api.u.pro.security.fix._common.plan.v1 import (  # noqa: F401
    AdditionalData,
    AptUpgradeData,
    AttachData,
    EnableData,
    FixPlanError,
    FixPlanResult,
    FixPlanStep,
    FixPlanUSNResult,
    FixPlanWarning,
    NoOpAlreadyFixedData,
    NoOpData,
    PackageCannotBeInstalledData,
    SecurityIssueNotFixedData,
    USNAdditionalData,
    fix_plan_usn,
)
from uaclient.config import UAConfig
from uaclient.data_types import DataObject, Field, StringDataValue, data_list


class USNFixPlanOptions(DataObject):
    fields = [
        Field(
            "usns",
            data_list(StringDataValue),
            doc="A list of USNs (i.e. USN-6119-1) titles",
        ),
    ]

    def __init__(self, usns: List[str]):
        self.usns = usns


class USNFixPlanResult(DataObject):
    fields = [
        Field(
            "expected_status",
            StringDataValue,
            doc="The expected status of fixing the USNs",
        ),
        Field(
            "usns",
            data_list(FixPlanUSNResult),
            doc="A list of ``FixPlanUSNResult`` objects",
        ),
    ]

    def __init__(self, *, expected_status: str, usns: List[FixPlanUSNResult]):
        self.expected_status = expected_status
        self.usns = usns


class USNSFixPlanResult(DataObject, AdditionalInfo):
    fields = [
        Field(
            "usns_data",
            USNFixPlanResult,
            doc="A list of ``USNFixPlanResult`` objects",
        ),
    ]

    def __init__(self, *, usns_data: USNFixPlanResult):
        self.usns_data = usns_data


def plan(options: USNFixPlanOptions) -> USNSFixPlanResult:
    return _plan(options, UAConfig())


def _plan(options: USNFixPlanOptions, cfg: UAConfig) -> USNSFixPlanResult:
    """
    This endpoint shows the necessary steps required to fix USNs in the system
    without executing any of those steps.
    """
    usns = []  # type: List[FixPlanUSNResult]
    expected_status = ""
    for usn in options.usns:
        usn_plan = fix_plan_usn(usn, cfg=cfg)
        expected_status = get_expected_overall_status(
            expected_status, usn_plan.target_usn_plan.expected_status
        )
        usns.append(usn_plan)

    return USNSFixPlanResult(
        usns_data=USNFixPlanResult(
            expected_status=expected_status,
            usns=usns,
        )
    )


endpoint = APIEndpoint(
    version="v1",
    name="USNFixPlan",
    fn=_plan,
    options_cls=USNFixPlanOptions,
)

_doc = {
    "introduced_in": "29",
    "requires_network": True,
    "example_python": """
from uaclient.api.u.pro.security.fix.usn.plan.v1 import plan, USNFixPlanOptions

options = USNFixPlanOptions(cves=["USN-1234-1", "USN-1235-1"])
result = plan(options)
""",  # noqa: E501
    "result_class": USNSFixPlanResult,
    "ignore_result_classes": [DataObject, AdditionalData],
    "extra_result_classes": [
        USNAdditionalData,
        AptUpgradeData,
        AttachData,
        EnableData,
        NoOpData,
        NoOpAlreadyFixedData,
        PackageCannotBeInstalledData,
        SecurityIssueNotFixedData,
    ],
    "exceptions": [],
    "example_cli": """pro api u.pro.security.fix.usn.plan.v1 --data '{"usns": ["USN-1234-1", "USN-1235-1"]}'""",  # noqa: E501
    "example_json": """
{
    "usns_data": {
        "expected_status": "fixed",
        "usns": [
            {
                "related_usns_plan": [],
                "target_usn_plan": {
                    "title": "USN-1234-5",
                    "expected_status": "fixed",
                    "plan": [
                        {
                            "operation": "apt-upgrade",
                            "order": 1,
                            "data": {
                                "binary_packages": ["pkg1"],
                                "source_packages": ["pkg1"],
                                "pocket": "standard-updates"
                            }
                        }
                    ],
                    "warnings": [],
                    "error": null,
                    "additional_data": {
                        "associated_cves": [
                            "CVE-1234-56789"
                        ],
                        "associated_launchpad_bus": [
                            "https://launchpad.net/bugs/BUG_ID"
                        ]
                    }
                },
            }
        ]
    }
}
""",
}



from typing import Optional  # noqa: F401

from uaclient import messages
from uaclient.entitlements import repo
from uaclient.types import (  # noqa: F401
    MessagingOperations,
    MessagingOperationsDict,
)

CC_README = "/usr/share/doc/ubuntu-commoncriteria/README"


class CommonCriteriaEntitlement(repo.RepoEntitlement):

    help_doc_url = messages.urls.COMMON_CRITERIA_HOME_PAGE
    name = "cc-eal"
    title = messages.CC_TITLE
    description = messages.CC_DESCRIPTION
    help_text = messages.CC_HELP_TEXT
    repo_key_file = "ubuntu-pro-cc-eal.gpg"
    apt_noninteractive = True
    supports_access_only = True
    origin = "UbuntuCC"

    @property
    def messaging(self) -> MessagingOperationsDict:
        post_enable = None  # type: Optional[MessagingOperations]
        if not self.access_only:
            post_enable = [messages.CC_POST_ENABLE.format(filename=CC_README)]
        return {
            "pre_install": [messages.CC_PRE_INSTALL],
            "post_enable": post_enable,
        }



import datetime
import logging
import os
import random
import re
import string
import sys
import tarfile
from typing import Dict, List, Optional, Tuple, Union  # noqa: F401

import paramiko
import yaml
from behave.model import Feature, Scenario
from behave.model_core import Status
from behave.runner import Context

import features.cloud as cloud
from features.util import (
    BUILDER_NAME_PREFIX,
    SUT,
    UA_TMP_DIR,
    InstallationSource,
    landscape_reject_all_pending_computers,
    process_template_vars,
)

PROCESS_LOG_TMPL = """\
returncode: {returncode}
stdout:
{stdout}
stderr:
{stderr}
"""


class UAClientBehaveConfig:
    """Store config options for Pro client behave test runs.

    This captures the configuration in one place, so that we have a single
    source of truth for test configuration (rather than having environment
    variable handling throughout the test code).

    :param cloud_credentials_path:
        Optional path the pycloudlib file containing the cloud credentials
    :param contract_token:
        A valid contract token to use during attach scenarios
    :param contract_token_legacy:
        A valid contract token with legacy support to use during attach
    :param contract_token_staging:
        A valid staging contract token to use during attach scenarios
    :param machine_types:
        A comma-separated string of machine_types to test: lxd-container,
            lxd-vm, azure.pro, azure.pro-fips, azure.generic, aws.pro,
            aws.pro-fips, aws.generic, gcp.pro, gcp.pro-fips, gcp.generic
    :param private_key_file:
        Optional path to pre-existing private key file to use when connecting
        launched VMs via ssh.
    :param private_key_name:
        Optional name of the cloud's named private key object to use when
        connecting to launched VMs via ssh. Default: uaclient-integration.
    :param destroy_instances:
        This boolean indicates that test containers should be destroyed after
        the completion. Set to False to leave instances running.
    :param debs_path:
        Location of the debs to be used when lauching a focal integration
        test. If that path is None, we will build those debs locally.
    :param artifact_dir:
        Location where test artifacts are emitted.
    """

    prefix = "UACLIENT_BEHAVE_"
    file_config_path = os.path.expanduser("~/.config/protest.yaml")

    # These variables are used in .from_environ() to convert the string
    # environment variable input to the appropriate Python types for use within
    # the test framework
    boolean_options = [
        "collect_coverage",
        "destroy_instances",
        "ephemeral_instance",
        "snapshot_strategy",
        "sbuild_output_to_terminal",
    ]
    str_options = [
        "cloud_credentials_path",
        "contract_token",
        "contract_token_legacy",
        "contract_token_staging",
        "contract_token_staging_expired",
        "contract_token_staging_expired_sometimes",
        "contract_staging_service_account_username",
        "contract_staging_service_account_password",
        "landscape_registration_key",
        "landscape_api_access_key",
        "landscape_api_secret_key",
        "machine_types",
        "releases",
        "private_key_file",
        "private_key_name",
        "artifact_dir",
        "install_from",
        "custom_ppa",
        "debs_path",
        "userdata_file",
        "check_version",
        "sbuild_chroot",
        "wsl_pubkey_path",
        "wsl_privkey_path",
        "wsl_ip_address",
    ]
    redact_options = [
        "contract_token",
        "contract_token_legacy",
        "contract_token_staging",
        "contract_token_staging_expired",
        "contract_token_staging_expired_sometimes",
        "contract_staging_service_account_password",
        "landscape_registration_key",
        "landscape_api_access_key",
        "landscape_api_secret_key",
    ]

    # This variable is used in .from_environ() but also to emit the "Config
    # options" stanza in __init__
    all_options = boolean_options + str_options

    def __init__(
        self,
        *,
        cloud_credentials_path: Optional[str] = None,
        collect_coverage: bool = False,
        destroy_instances: bool = True,
        ephemeral_instance: bool = False,
        snapshot_strategy: bool = False,
        sbuild_output_to_terminal: bool = False,
        machine_types: Optional[str] = None,
        releases: Optional[str] = None,
        private_key_file: Optional[str] = None,
        private_key_name: str = "uaclient-integration",
        contract_token: Optional[str] = None,
        contract_token_legacy: Optional[str] = None,
        contract_token_staging: Optional[str] = None,
        contract_token_staging_expired: Optional[str] = None,
        contract_token_staging_expired_sometimes: Optional[str] = None,
        contract_staging_service_account_username: Optional[str] = None,
        contract_staging_service_account_password: Optional[str] = None,
        landscape_registration_key: Optional[str] = None,
        landscape_api_access_key: Optional[str] = None,
        landscape_api_secret_key: Optional[str] = None,
        artifact_dir: str = "artifacts",
        install_from: InstallationSource = InstallationSource.LOCAL,
        custom_ppa: Optional[str] = None,
        debs_path: Optional[str] = None,
        userdata_file: Optional[str] = None,
        check_version: Optional[str] = None,
        sbuild_chroot: Optional[str] = None,
        wsl_pubkey_path: Optional[str] = None,
        wsl_privkey_path: Optional[str] = None,
        wsl_ip_address: Optional[str] = None,
    ) -> None:
        # First, store the values we've detected
        self.cloud_credentials_path = cloud_credentials_path
        self.collect_coverage = collect_coverage
        self.ephemeral_instance = ephemeral_instance
        self.snapshot_strategy = snapshot_strategy
        self.sbuild_output_to_terminal = sbuild_output_to_terminal
        self.contract_token = contract_token
        self.contract_token_legacy = contract_token_legacy
        self.contract_token_staging = contract_token_staging
        self.contract_token_staging_expired = contract_token_staging_expired
        self.contract_token_staging_expired_sometimes = (
            contract_token_staging_expired_sometimes
        )
        self.contract_staging_service_account_username = (
            contract_staging_service_account_username
        )
        self.contract_staging_service_account_password = (
            contract_staging_service_account_password
        )
        self.landscape_registration_key = landscape_registration_key
        self.landscape_api_access_key = landscape_api_access_key
        self.landscape_api_secret_key = landscape_api_secret_key
        self.destroy_instances = destroy_instances
        self.private_key_file = private_key_file
        self.private_key_name = private_key_name
        self.artifact_dir = artifact_dir
        self.install_from = install_from
        self.custom_ppa = custom_ppa
        self.debs_path = debs_path
        self.userdata_file = userdata_file
        self.check_version = check_version
        self.sbuild_chroot = sbuild_chroot
        self.wsl_pubkey_path = wsl_pubkey_path
        self.wsl_privkey_path = wsl_privkey_path
        self.wsl_ip_address = wsl_ip_address

        self.machine_types = (
            machine_types.split(",") if machine_types else None
        )
        self.releases = releases.split(",") if releases else None
        # Next, perform any required validation
        if install_from == InstallationSource.CUSTOM and not custom_ppa:
            logging.error(
                "UACLIENT_BEHAVE_INSTALL_FROM is set to 'custom', "
                "but missing UACLIENT_BEHAVE_CUSTOM_PPA"
            )
            sys.exit(1)
        if install_from == InstallationSource.PREBUILT and not debs_path:
            logging.error(
                "UACLIENT_BEHAVE_INSTALL_FROM is set to 'prebuilt', "
                "but missing UACLIENT_BEHAVE_DEBS_PATH"
            )
            sys.exit(1)

        ignore_vars = ()  # type: Tuple[str, ...]
        if (
            self.machine_types
            and len(self.machine_types) == 1
            and "pro" in self.machine_types[0]
        ):
            ignore_vars += (
                "UACLIENT_BEHAVE_CONTRACT_TOKEN",
                "UACLIENT_BEHAVE_CONTRACT_TOKEN_STAGING",
                "UACLIENT_BEHAVE_CONTRACT_TOKEN_STAGING_EXPIRED",
            )
        for env_name in ignore_vars:
            attr_name = env_name.replace("UACLIENT_BEHAVE_", "").lower()
            if getattr(self, attr_name):
                print(
                    " --- Ignoring {} because machine_types is {}".format(
                        env_name, self.machine_types
                    )
                )
                setattr(self, attr_name, None)
        timed_job_tag = datetime.datetime.utcnow().strftime(
            "uaclient-ci-%m%d-%H%M-"
        )
        # Jenkinsfile provides us with UACLIENT_BEHAVE_JENKINS_BUILD_TAG
        job_suffix = os.environ.get("UACLIENT_BEHAVE_JENKINS_BUILD_TAG")
        print("--- job suffix: {}".format(job_suffix), flush=True)
        if not job_suffix:
            job_suffix = os.environ.get("CHANGE_ID", "dev")
        else:
            job_suffix = job_suffix.split("PR-")[-1]
        timed_job_tag += str(job_suffix)
        timed_job_tag = timed_job_tag.replace(".", "-")

        # Add 8-digit random suffix to the tag, so it does not conflict when
        # spinning many jobs at the same time
        random_suffix = "".join(
            random.choices(string.ascii_lowercase + string.digits, k=8)
        )
        timed_job_tag += "-" + random_suffix
        self.timed_job_tag = timed_job_tag

        self.clouds = cloud.CloudManager(self)

        # Finally, print the config options.  This helps users debug the use of
        # config options, and means they'll be included in test logs in CI.
        print("Config options:")
        for option in self.all_options:
            value = getattr(self, option, "<UNSET>")
            if option in self.redact_options and value not in (
                None,
                "<UNSET>",
            ):
                value = "<REDACTED>"
            print("  {}".format(option), "=", value)

    @classmethod
    def from_environ(cls, config) -> "UAClientBehaveConfig":
        """Gather config options from:

        1. ~/.config/protest.yaml
        2. os.environ with prefix UACLIENT_BEHAVE_
        3. -D command line options

        each of which can override the previous.
        """
        # First, gather all known options
        kwargs = (
            {}
        )  # type: Dict[str, Union[str, bool, List, InstallationSource]]

        try:
            with open(cls.file_config_path) as file_config:
                kwargs = yaml.safe_load(file_config)
        except FileNotFoundError:
            print("No config file found at {}".format(cls.file_config_path))

        for key, value in os.environ.items():
            if not key.startswith(cls.prefix):
                continue
            our_key = key[len(cls.prefix) :].lower()
            if our_key not in cls.all_options:
                print("Unknown config environment variable:", key)
                continue
            kwargs[our_key] = value

        # userdata should override environment variables
        kwargs.update(config.userdata)

        # Next, sanitise the non-string options to Python types
        for key in cls.boolean_options:
            bool_value = True  # Default to True
            if key in kwargs:
                if kwargs[key] == "0" or str(kwargs[key]).lower() == "false":
                    bool_value = False
                kwargs[key] = bool_value

        if "install_from" in kwargs:
            if str(kwargs["install_from"]).startswith("ppa:"):
                kwargs["custom_ppa"] = kwargs["install_from"]
                kwargs["install_from"] = InstallationSource.CUSTOM
            else:
                kwargs["install_from"] = InstallationSource(
                    kwargs["install_from"]
                )

        return cls(**kwargs)  # type: ignore


def before_all(context: Context) -> None:
    """behave will invoke this before anything else happens."""
    context.config.setup_logging()
    if logging.getLogger().level == logging.DEBUG:
        # The AWS boto libraries are very very very verbose
        # We pretty much never want their debug level messages,
        # but we do want to be able to use the debug log level
        # in our code. So we bump their loggers up to info
        # when we set debug at the cli with --logging-level=debug
        logging.warn(
            "Setting AWS botocore and boto3 loggers to INFO to avoid"
            " extra verbose logs"
        )
        logging.getLogger("botocore").setLevel(logging.INFO)
        logging.getLogger("boto3").setLevel(logging.INFO)
    userdata = context.config.userdata
    if userdata:
        logging.debug("Userdata key / value pairs:")
        print("Userdata key / value pairs:")
        for key, value in userdata.items():
            logging.debug("   - {} = {}".format(key, value))
            print("   - {} = {}".format(key, value))
    context.pro_config = UAClientBehaveConfig.from_environ(context.config)
    context.snapshots = {}
    context.machines = {}


def _should_skip_config_tags(context: Context, tags: List) -> str:
    """
    Return a reason if a feature or scenario should be skipped based on
    missing but required config tags
    """

    for tag in tags:
        parts = tag.split(".")
        if parts[0] != "uses" or parts[1] != "config":
            continue  # Only process @uses.config.* tags for skipping:
        val = context.pro_config
        for idx, attr in enumerate(parts[2:], 1):
            val = getattr(val, attr, None)
            if val is None:
                return "Skipped: tag value was None: {}".format(tag)

    return ""


def before_feature(context: Context, feature: Feature):
    reason = _should_skip_config_tags(context, feature.tags)
    if reason:
        feature.skip(reason=reason)


def before_scenario(context: Context, scenario: Scenario):
    context.stored_vars = {}

    reason = _should_skip_config_tags(context, scenario.effective_tags)
    if reason:
        scenario.skip(reason=reason)
        return

    # Determine release and machine_type of this scenario.
    # First check outline example table row.
    # Then override with what is passed directly to the "Give a machine" step
    # if applicable.
    scenario_release = None
    scenario_machine_type = None

    if hasattr(scenario, "_row") and scenario._row is not None:
        scenario_release = scenario._row.get("release")
        scenario_machine_type = scenario._row.get("machine_type")

    given_a_series_machine_type_match = re.match(
        "a `(.*)` `(.*)` machine with ubuntu-advantage-tools installed",
        scenario.steps[0].name,
    )
    if given_a_series_machine_type_match:
        step_release = given_a_series_machine_type_match.group(1)
        if step_release != "<release>":
            scenario_release = step_release
        step_machine_type = given_a_series_machine_type_match.group(2)
        if step_machine_type != "<machine_type>":
            scenario_machine_type = step_machine_type

    releases = context.pro_config.releases
    if releases and scenario_release not in releases:
        scenario.skip(
            reason=(
                "Scenario release is `{}`, but releases filter set"
                " to {} - skipping.".format(scenario_release, releases)
            )
        )
        return
    machine_types = context.pro_config.machine_types
    if machine_types and scenario_machine_type not in machine_types:
        scenario.skip(
            reason=(
                "Scenario machine_type is `{}`, but machine_types filter set"
                " to {} - skipping.".format(
                    scenario_machine_type, context.pro_config.machine_types
                )
            )
        )
        return

    install_from = context.pro_config.install_from
    if install_from == InstallationSource.LOCAL:
        if "skip_local_environment" in scenario.effective_tags:
            scenario.skip(
                reason="Scenario does not support install_from local"
            )
            return

    if install_from == InstallationSource.PREBUILT:
        if "skip_prebuilt_environment" in scenario.effective_tags:
            scenario.skip(
                reason="Scenario does not support install_from prebuilt"
            )
            return

    # before_step doesn't execute early enough to modify the step
    # so we perform step text surgery here
    # Also, logging capture is not set up when before_scenario is called,
    # so if you call logging.info here, then the root logger gets configured
    # and it messes up all the future behave logging capture machinery.
    # See https://github.com/behave/behave/blob/v1.2.6/behave/model.py#L700
    # But we want to log the replacement we're making, so we use the behave
    # logger and warning log_level to make sure it gets included.
    logger = logging.getLogger("behave.before_scenario.process_template_vars")
    for step in scenario.steps:
        if step.text:
            step.text = process_template_vars(
                context, step.text, logger_fn=logger.warn, shown=True
            )


def after_scenario(context, scenario):
    """Collect the coverage files after the scenario is run."""
    if context.pro_config.collect_coverage:
        cov_dir = os.path.join(context.pro_config.artifact_dir, "coverage")
        os.makedirs(cov_dir, exist_ok=True)

        inner_dir = os.path.basename(scenario.filename.replace(".", "_"))
        new_artifacts_dir = os.path.join(
            cov_dir,
            inner_dir,
        )
        os.makedirs(new_artifacts_dir, exist_ok=True)
        if hasattr(context, "machines") and SUT in context.machines:
            try:
                scenario_name = (
                    os.path.basename(scenario.filename.replace(".", "_"))
                    + "_"
                    + str(scenario.line)
                )
                cov_filename = ".coverage.{}".format(scenario_name)
                tmp_covfile_path = "/tmp/{cov_filename}".format(
                    cov_filename=cov_filename
                )
                context.machines[SUT].instance.execute(
                    [
                        "mv",
                        "/home/ubuntu/.coverage",
                        tmp_covfile_path,
                    ],
                    use_sudo=True,
                )
                context.machines[SUT].instance.execute(
                    [
                        "chmod",
                        "666",
                        tmp_covfile_path,
                    ],
                    use_sudo=True,
                )

                dest = os.path.join(new_artifacts_dir, cov_filename)
                context.machines[SUT].instance.pull_file(
                    tmp_covfile_path, dest
                )
                logging.warning("Done collecting coverage.")
            except Exception as e:
                logging.error(str(e))
                logging.warning("Failed to collect coverage")


def _get_relevant_apparmor_logs(context):
    if hasattr(context, "machines") and SUT in context.machines:
        sut = context.machines[SUT]
        if sut.cloud == "lxd-container":
            # get apparmor DENIED messages from the host
            with open("/var/log/syslog", "r") as syslog_fd:
                syslog_messages = syslog_fd.readlines()
            apparmor_denied = [
                msg.strip()
                for msg in syslog_messages
                if (
                    "DENIED" in msg
                    and "ubuntu_pro_" in msg
                    and sut.instance.name in msg
                )
            ]
            return apparmor_denied
        else:
            # get apparmor DENIED messages from the SUT
            os.makedirs(UA_TMP_DIR, exist_ok=True)
            syslog_dest = os.path.join(
                UA_TMP_DIR, "{}-syslog".format(sut.instance.name)
            )
            try:
                sut.instance.pull_file("/var/log/syslog", syslog_dest)
            except paramiko.ssh_exception.SSHException:
                logging.warning(
                    "Unable to pull syslog. Skipping apparmor log check."
                )
                return None
            except FileNotFoundError:
                logging.warning(
                    "syslog file doesn't exist. Skipping apparmor log check."
                )
                return None

            with open(syslog_dest, "r") as syslog_fd:
                syslog_messages = syslog_fd.readlines()
            apparmor_denied = [
                msg.strip()
                for msg in syslog_messages
                if ("DENIED" in msg and "ubuntu_pro_" in msg)
            ]
            return apparmor_denied
    return None


def after_step(context, step):
    """Collect test artifacts in the event of failure."""
    apparmor_logs = _get_relevant_apparmor_logs(context)
    if apparmor_logs:
        logging.warning("XXX apparmor DENIED begin")
        logging.warning("\n".join(apparmor_logs))
        logging.warning("XXX apparmor DENIED end")
        # naughty
        step.status = Status.failed
    if step.status == "failed":
        logging.warning("STEP FAILED. Collecting logs.")
        inner_dir = os.path.join(
            datetime.datetime.now().strftime("%Y-%m-%dT%H-%M-%S"),
            "{}_{}".format(os.path.basename(step.filename), step.line),
        )
        new_artifacts_dir = os.path.join(
            context.pro_config.artifact_dir,
            inner_dir,
        )
        if not os.path.exists(new_artifacts_dir):
            os.makedirs(new_artifacts_dir)

        latest_link_dir = os.path.join(
            context.pro_config.artifact_dir, "latest"
        )
        if os.path.exists(latest_link_dir):
            os.unlink(latest_link_dir)
        os.symlink(inner_dir, latest_link_dir)

        if hasattr(context, "process"):
            artifact_file = os.path.join(new_artifacts_dir, "process.log")
            process = context.process
            with open(artifact_file, "w") as stream:
                stream.write(
                    PROCESS_LOG_TMPL.format(
                        returncode=process.returncode,
                        stdout=process.stdout,
                        stderr=process.stderr,
                    )
                )

        if hasattr(context, "machines") and SUT in context.machines:
            try:
                context.machines[SUT].instance.execute(
                    ["pro", "collect-logs", "-o", "/tmp/logs.tar.gz"],
                    use_sudo=True,
                )
                context.machines[SUT].instance.execute(
                    ["chmod", "666", "/tmp/logs.tar.gz"], use_sudo=True
                )
                dest = os.path.join(new_artifacts_dir, "logs.tar.gz")
                context.machines[SUT].instance.pull_file(
                    "/tmp/logs.tar.gz", dest
                )
                with tarfile.open(dest) as logs_tarfile:
                    logs_tarfile.extractall(new_artifacts_dir)
                logging.warning("Done collecting logs.")
            except Exception as e:
                logging.error(str(e))
                logging.warning("Failed to collect logs")


def after_all(context):
    if context.pro_config.destroy_instances:
        try:
            for cloud_instance in context.pro_config.clouds.clouds.values():
                if cloud_instance._ssh_key_managed:
                    key_pair = cloud_instance.api.key_pair
                    os.remove(key_pair.private_key_path)
                    os.remove(key_pair.public_key_path)

        except Exception as e:
            logging.error(
                "Failed to delete instance ssh keys:\n{}".format(str(e))
            )

        if context.pro_config.clouds.has("wsl"):
            cloud_instance = context.pro_config.clouds.get("wsl")
            cloud_instance.stop_windows_machine()

    # Builder snapshots don't get an auto-cleanup function, so clean them here
    builder_snapshots = [
        name
        for name in context.snapshots
        if name.startswith(BUILDER_NAME_PREFIX)
    ]
    for snapshot in builder_snapshots:
        cloud = context.snapshots[snapshot].cloud
        try:
            context.pro_config.clouds.get(cloud).api.delete_image(
                context.snapshots[snapshot].name
            )
        except RuntimeError as e:
            logging.error(
                "Failed to delete image: {}\n{}".format(
                    context.snapshots[snapshot].name, str(e)
                )
            )

    if (
        context.pro_config.landscape_api_access_key
        and context.pro_config.landscape_api_secret_key
    ):
        landscape_reject_all_pending_computers(context)



import json
import logging
import re

import mock
import pytest

from uaclient import exceptions, http
from uaclient.clouds.aws import (
    AWS_TOKEN_PUT_HEADER,
    AWS_TOKEN_REQ_HEADER,
    AWS_TOKEN_TTL_SECONDS,
    IMDS_IPV4_ADDRESS,
    IMDS_IPV6_ADDRESS,
    IMDS_URL,
    IMDS_V2_TOKEN_URL,
    AWSAutoAttachInstance,
)

M_PATH = "uaclient.clouds.aws."


class TestAWSAutoAttachInstance:
    def test_cloud_type(self):
        instance = AWSAutoAttachInstance()
        assert "aws" == instance.cloud_type

    @mock.patch(M_PATH + "http.readurl")
    def test__get_imds_v2_token_headers_none_on_404(self, readurl):
        """A 404 on private AWS regions indicates lack IMDSv2 support."""
        readurl.return_value = http.HTTPResponse(
            code=404,
            headers={},
            body="",
            json_dict={},
            json_list=[],
        )
        instance = AWSAutoAttachInstance()
        assert None is instance._get_imds_v2_token_headers(
            ip_address=IMDS_IPV4_ADDRESS
        )
        assert "IMDSv1" == instance._api_token
        # No retries on 404. It is a permanent indication of no IMDSv2 support.
        instance._get_imds_v2_token_headers(ip_address=IMDS_IPV4_ADDRESS)
        assert 1 == readurl.call_count

    @mock.patch(M_PATH + "http.readurl")
    def test__get_imds_v2_token_headers_caches_response(self, readurl):
        """Return API token headers for IMDSv2 access. Response is cached."""
        instance = AWSAutoAttachInstance()
        url = "http://169.254.169.254/latest/api/token"
        readurl.return_value = http.HTTPResponse(
            code=200,
            headers={"header": "stuff"},
            body="somebase64token==",
            json_dict={},
            json_list=[],
        )
        assert {
            AWS_TOKEN_PUT_HEADER: "somebase64token=="
        } == instance._get_imds_v2_token_headers(ip_address=IMDS_IPV4_ADDRESS)
        instance._get_imds_v2_token_headers(ip_address=IMDS_IPV4_ADDRESS)
        assert "somebase64token==" == instance._api_token
        assert [
            mock.call(
                url,
                method="PUT",
                headers={AWS_TOKEN_REQ_HEADER: AWS_TOKEN_TTL_SECONDS},
                timeout=1,
            )
        ] == readurl.call_args_list

    @pytest.mark.parametrize("caplog_text", [logging.DEBUG], indirect=True)
    @pytest.mark.parametrize("fail_count,exception", ((3, False), (4, True)))
    @mock.patch(M_PATH + "util.time.sleep")
    @mock.patch(M_PATH + "http.readurl")
    def test_retry_backoff_on__get_imds_v2_token_headers_caches_response(
        self, readurl, sleep, fail_count, exception, caplog_text
    ):
        """Retry backoff before failing _get_imds_v2_token_headers."""

        def fake_someurlerrors(url, method=None, headers=None, timeout=1):
            if readurl.call_count <= fail_count:
                return http.HTTPResponse(
                    code=700 + readurl.call_count,
                    headers={},
                    body="funky error msg",
                    json_dict={},
                    json_list=[],
                )
            return http.HTTPResponse(
                code=200,
                headers={"header": "stuff"},
                body="base64token==",
                json_dict={},
                json_list=[],
            )

        readurl.side_effect = fake_someurlerrors
        instance = AWSAutoAttachInstance()
        if exception:
            with pytest.raises(exceptions.CloudMetadataError):
                instance._get_imds_v2_token_headers(
                    ip_address=IMDS_IPV4_ADDRESS
                )
        else:
            assert {
                AWS_TOKEN_PUT_HEADER: "base64token=="
            } == instance._get_imds_v2_token_headers(
                ip_address=IMDS_IPV4_ADDRESS
            )

        expected_sleep_calls = [mock.call(1), mock.call(2), mock.call(5)]
        assert expected_sleep_calls == sleep.call_args_list
        expected_logs = [
            "An error occurred while talking the the cloud metadata service: 701 - funky error msg: Retrying 3 more times.",  # noqa: E501
            "An error occurred while talking the the cloud metadata service: 702 - funky error msg: Retrying 2 more times.",  # noqa: E501
            "An error occurred while talking the the cloud metadata service: 703 - funky error msg: Retrying 1 more times.",  # noqa: E501
        ]
        logs = caplog_text()
        for log in expected_logs:
            assert log in logs

    @mock.patch(M_PATH + "http.readurl")
    def test_identity_doc_from_aws_url_pkcs7(self, readurl):
        """Return pkcs7 content from IMDS as AWS' identity doc"""
        readurl.return_value = http.HTTPResponse(
            code=200,
            headers={"header": "stuff"},
            body="pkcs7WOOT!==",
            json_dict={},
            json_list=[],
        )
        instance = AWSAutoAttachInstance()
        assert {"pkcs7": "pkcs7WOOT!=="} == instance.identity_doc
        url = "http://169.254.169.254/latest/dynamic/instance-identity/pkcs7"
        token_url = "http://169.254.169.254/latest/api/token"
        assert [
            mock.call(
                token_url,
                method="PUT",
                headers={AWS_TOKEN_REQ_HEADER: AWS_TOKEN_TTL_SECONDS},
                timeout=1,
            ),
            mock.call(
                url, headers={AWS_TOKEN_PUT_HEADER: "pkcs7WOOT!=="}, timeout=1
            ),
        ] == readurl.call_args_list

    @pytest.mark.parametrize("caplog_text", [logging.DEBUG], indirect=True)
    @pytest.mark.parametrize("fail_count,exception", ((3, False),))
    @mock.patch(M_PATH + "util.time.sleep")
    @mock.patch(M_PATH + "http.readurl")
    def test_retry_backoff_on_failed_identity_doc(
        self, readurl, sleep, fail_count, exception, caplog_text
    ):
        """Retry backoff is attempted before failing to get AWS.identity_doc"""

        def fake_someurlerrors(url, method=None, headers=None, timeout=1):
            # due to _get_imds_v2_token_headers
            if "latest/api/token" in url:
                return http.HTTPResponse(
                    code=200,
                    headers={"header": "stuff"},
                    body="base64token==",
                    json_dict={},
                    json_list=[],
                )
            if readurl.call_count <= fail_count + 1:
                return http.HTTPResponse(
                    code=700 + readurl.call_count,
                    headers={},
                    body="funky error msg",
                    json_dict={},
                    json_list=[],
                )
            return http.HTTPResponse(
                code=200,
                headers={"header": "stuff"},
                body="pkcs7WOOT!==",
                json_dict={},
                json_list=[],
            )

        readurl.side_effect = fake_someurlerrors
        instance = AWSAutoAttachInstance()
        if exception:
            with pytest.raises(exceptions.CloudMetadataError) as excinfo:
                instance.identity_doc
            assert 705 == excinfo.value.code
        else:
            assert {"pkcs7": "pkcs7WOOT!=="} == instance.identity_doc

        expected_sleep_calls = [mock.call(0.5), mock.call(1), mock.call(1)]
        assert expected_sleep_calls == sleep.call_args_list
        expected_logs = [
            "An error occurred while talking the the cloud metadata service: 702 - funky error msg: Retrying 3 more times.",  # noqa: E501
            "An error occurred while talking the the cloud metadata service: 703 - funky error msg: Retrying 2 more times.",  # noqa: E501
            "An error occurred while talking the the cloud metadata service: 704 - funky error msg: Retrying 1 more times.",  # noqa: E501
        ]
        logs = caplog_text()
        for log in expected_logs:
            assert log in logs

    @pytest.mark.parametrize("uuid", ("ec2", "ec2yep"))
    @mock.patch(M_PATH + "system.load_file")
    def test_is_viable_based_on_sys_hypervisor_uuid(self, load_file, uuid):
        """Viable ec2 platform is determined by /sys/hypervisor/uuid prefix"""
        load_file.return_value = uuid
        instance = AWSAutoAttachInstance()
        assert True is instance.is_viable

    @pytest.mark.parametrize(
        "hypervisor_uuid,prod_uuid,prod_serial,viable",
        (
            ("notec2", "ec2UUID", "ec2Serial", True),
            (None, "ec2UUID", "ec2Serial", True),
            ("notec2", "EC2UUID", "Ec2Serial", True),
            (None, "EC2UUID", "Ec2Serial", True),
            ("notec2", "!EC2UUID", "Ec2Serial", False),
            ("notec2", "EC2UUID", "1Ec2Serial", False),
            ("notec2", "ec2UUID", "ec3Serial", False),
            ("notec2", "ec3UUID", "ec2Serial", False),
        ),
    )
    @mock.patch(M_PATH + "system.load_file")
    def test_is_viable_based_on_sys_product_serial_and_uuid(
        self, load_file, hypervisor_uuid, prod_uuid, prod_serial, viable
    ):
        """Platform is viable when product serial and uuid start with ec2"""

        def fake_load_file(f_name):
            if f_name == "/sys/hypervisor/uuid":
                if hypervisor_uuid is not None:
                    return hypervisor_uuid
                raise FileNotFoundError()
            if f_name == "/sys/class/dmi/id/product_uuid":
                return prod_uuid
            if f_name == "/sys/class/dmi/id/product_serial":
                return prod_serial
            raise AssertionError("Invalid load_file of {}".format(f_name))

        load_file.side_effect = fake_load_file
        instance = AWSAutoAttachInstance()
        assert viable is instance.is_viable

    @pytest.mark.parametrize("caplog_text", [logging.DEBUG], indirect=True)
    @mock.patch(M_PATH + "http.readurl")
    def test_identity_doc_default_to_ipv6_if_ipv4_fail(
        self, readurl, caplog_text
    ):
        instance = AWSAutoAttachInstance()
        ipv4_address = IMDS_IPV4_ADDRESS
        ipv6_address = IMDS_IPV6_ADDRESS

        def fake_someurlerrors(url, method=None, headers=None, timeout=1):
            if ipv4_address in url:
                raise Exception("IPv4 exception")

            if url == IMDS_V2_TOKEN_URL.format(ipv6_address):
                return http.HTTPResponse(
                    code=200,
                    headers={"header": "stuff"},
                    body="base64token==",
                    json_dict={},
                    json_list=[],
                )

            if url == IMDS_URL.format(ipv6_address):
                return http.HTTPResponse(
                    code=200,
                    headers={"header": "stuff"},
                    body="pkcs7WOOT!==",
                    json_dict={},
                    json_list=[],
                )

        readurl.side_effect = fake_someurlerrors
        assert {"pkcs7": "pkcs7WOOT!=="} == instance.identity_doc
        expected = [
            mock.call(
                IMDS_V2_TOKEN_URL.format(ipv4_address),
                method="PUT",
                headers={AWS_TOKEN_REQ_HEADER: AWS_TOKEN_TTL_SECONDS},
                timeout=1,
            ),
            mock.call(
                IMDS_V2_TOKEN_URL.format(ipv6_address),
                method="PUT",
                headers={AWS_TOKEN_REQ_HEADER: AWS_TOKEN_TTL_SECONDS},
                timeout=1,
            ),
            mock.call(
                IMDS_URL.format(ipv6_address),
                headers={AWS_TOKEN_PUT_HEADER: "base64token=="},
                timeout=1,
            ),
        ]

        assert expected == readurl.call_args_list

        expected_log = "Could not reach AWS IMDS at http://169.254.169.254:"
        assert expected_log in caplog_text()

    @pytest.mark.parametrize("caplog_text", [logging.DEBUG], indirect=True)
    @mock.patch(M_PATH + "http.readurl")
    def test_identity_doc_logs_error_if_both_ipv4_and_ipv6_fails(
        self, readurl, caplog_text
    ):

        instance = AWSAutoAttachInstance()
        ipv4_address = IMDS_IPV4_ADDRESS
        ipv6_address = IMDS_IPV6_ADDRESS

        readurl.side_effect = Exception("Exception")

        expected_error = (
            "No valid AWS IMDS endpoint discovered at "
            "addresses: {}, {}".format(IMDS_IPV4_ADDRESS, IMDS_IPV6_ADDRESS)
        )
        with pytest.raises(
            exceptions.UbuntuProError, match=re.escape(expected_error)
        ):
            instance.identity_doc

        expected = [
            mock.call(
                IMDS_V2_TOKEN_URL.format(ipv4_address),
                method="PUT",
                headers={AWS_TOKEN_REQ_HEADER: AWS_TOKEN_TTL_SECONDS},
                timeout=1,
            ),
            mock.call(
                IMDS_V2_TOKEN_URL.format(ipv6_address),
                method="PUT",
                headers={AWS_TOKEN_REQ_HEADER: AWS_TOKEN_TTL_SECONDS},
                timeout=1,
            ),
        ]
        assert expected == readurl.call_args_list

        expected_logs = [
            "Could not reach AWS IMDS at http://169.254.169.254:",
            "Could not reach AWS IMDS at http://[fd00:ec2::254]:",
        ]

        for expected_log in expected_logs:
            assert expected_log in caplog_text()

    def test_unsupported_should_poll_for_pro_license(self):
        """Unsupported"""
        instance = AWSAutoAttachInstance()
        assert not instance.should_poll_for_pro_license()

    def test_unsupported_is_pro_license_present(self):
        """Unsupported"""
        instance = AWSAutoAttachInstance()
        with pytest.raises(exceptions.InPlaceUpgradeNotSupportedError):
            instance.is_pro_license_present(wait_for_change=False)

    def test__get_ii_doc_ok(self):
        instance = AWSAutoAttachInstance()
        instance._ip_address = "169.254.169.254"
        headers = {"X-aws-ec2-metadata-token": "token"}
        instance._request_imds_v2_token_headers = mock.MagicMock(
            return_value=headers
        )

        ii_doc = {
            "accountId": "937157663530",
            "architecture": "x86_64",
            "availabilityZone": "eu-south-2b",
            "billingProducts": ["bp-66a5400f"],
            "devpayProductCodes": None,
            "marketplaceProductCodes": None,
            "imageId": "ami-0691695a958575df9",
            "instanceId": "i-09d1747b6cf64cdb9",
            "instanceType": "t3.micro",
            "kernelId": None,
            "pendingTime": "2024-08-26T09:27:51Z",
            "privateIp": "192.168.2.157",
            "ramdiskId": None,
            "region": "eu-south-2",
            "version": "2017-09-30",
        }
        instance._get_imds_url_response = mock.MagicMock(
            return_value=json.dumps(ii_doc)
        )

        assert ii_doc == instance._get_ii_doc()

        assert [
            mock.call(
                "http://169.254.169.254/latest/dynamic/instance-identity"
                "/document",
                headers=headers,
            ),
        ] == instance._get_imds_url_response.call_args_list

    def test__get_ii_doc_json_error(self):
        """test behavior when json.load fails"""
        instance = AWSAutoAttachInstance()
        instance._ip_address = "169.254.169.254"
        headers = {"X-aws-ec2-metadata-token": "token"}
        instance._request_imds_v2_token_headers = mock.MagicMock(
            return_value=headers
        )
        instance._get_imds_url_response = mock.MagicMock(
            return_value="{invalidjson"
        )
        assert {} == instance._get_ii_doc()
        assert [
            mock.call(
                "http://169.254.169.254/latest/dynamic/instance-identity"
                "/document",
                headers=headers,
            ),
        ] == instance._get_imds_url_response.call_args_list

    @pytest.mark.parametrize(
        "ii_doc, is_likely_pro",
        [
            (
                {
                    "accountId": "937157663530",
                    "architecture": "x86_64",
                    "availabilityZone": "eu-south-2b",
                    "billingProducts": ["bp-66a5400f"],
                    "devpayProductCodes": None,
                    "marketplaceProductCodes": None,
                    "imageId": "ami-0691695a958575df9",
                    "instanceId": "i-09d1747b6cf64cdb9",
                    "instanceType": "t3.micro",
                    "kernelId": None,
                    "pendingTime": "2024-08-26T09:27:51Z",
                    "privateIp": "192.168.2.157",
                    "ramdiskId": None,
                    "region": "eu-south-2",
                    "version": "2017-09-30",
                },
                True,
            ),
            (
                {
                    "accountId": "937157663530",
                    "architecture": "x86_64",
                    "availabilityZone": "eu-south-2b",
                    "billingProducts": ["bp-66a5400f"],
                    "devpayProductCodes": None,
                    "marketplaceProductCodes": ["mp-66a5400f"],
                    "imageId": "ami-0691695a958575df9",
                    "instanceId": "i-09d1747b6cf64cdb9",
                    "instanceType": "t3.micro",
                    "kernelId": None,
                    "pendingTime": "2024-08-26T09:27:51Z",
                    "privateIp": "192.168.2.157",
                    "ramdiskId": None,
                    "region": "eu-south-2",
                    "version": "2017-09-30",
                },
                True,
            ),
            (
                {
                    "accountId": "937157663530",
                    "architecture": "x86_64",
                    "availabilityZone": "eu-south-2b",
                    "billingProducts": None,
                    "devpayProductCodes": None,
                    "marketplaceProductCodes": ["mp-66a5400f"],
                    "imageId": "ami-0691695a958575df9",
                    "instanceId": "i-09d1747b6cf64cdb9",
                    "instanceType": "t3.micro",
                    "kernelId": None,
                    "pendingTime": "2024-08-26T09:27:51Z",
                    "privateIp": "192.168.2.157",
                    "ramdiskId": None,
                    "region": "eu-south-2",
                    "version": "2017-09-30",
                },
                True,
            ),
            (
                {
                    "billingProducts": None,
                    "marketplaceProductCodes": ["mp-66a5400f"],
                },
                True,
            ),
            (
                {
                    "billingProducts": None,
                    "marketplaceProductCodes": None,
                },
                False,
            ),
            (
                {
                    "billingProducts": [],
                    "marketplaceProductCodes": [],
                },
                False,
            ),
            ({}, False),
        ],
    )
    def test_is_likely_pro(self, ii_doc, is_likely_pro):
        instance = AWSAutoAttachInstance()
        instance._get_ii_doc = mock.MagicMock(return_value=ii_doc)
        assert is_likely_pro == instance.is_likely_pro
